"""
YTreeArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import json
import numpy as np
import os

from unyt.unit_registry import \
    UnitRegistry

from yt.data_objects.data_containers import \
    YTDataContainer
from yt.utilities.logger import \
    ytLogger
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_objects

from ytree.data_structures.arbor import \
    Arbor
from ytree.data_structures.load import load as ytree_load
from ytree.frontends.ytree.io import \
    YTreeDataFile, \
    YTreeRootFieldIO, \
    YTreeTreeFieldIO
from ytree.frontends.ytree.utilities import \
    get_about, \
    get_conditional
from ytree.utilities.io import \
    _hdf5_yt_attr, \
    parse_h5_attr
from ytree.utilities.logger import \
    log_level
from ytree.yt_frontend import \
    YTreeDataset

class YTreeArbor(Arbor):
    """
    Class for Arbors created from the
    :func:`~ytree.data_structures.arbor.Arbor.save_arbor`
    or :func:`~ytree.data_structures.tree_node.TreeNode.save_tree` functions.
    """
    _root_field_io_class = YTreeRootFieldIO
    _tree_field_io_class = YTreeTreeFieldIO
    _suffix = ".h5"
    _node_io_attrs = ('_ai',)

    def _node_io_loop_prepare(self, nodes):
        if nodes is None:
            nodes = np.arange(self.size)
            ai = self._node_info['_ai']
        elif nodes.dtype == object:
            ai = np.array(
                [node._ai if node.is_root else node.root._ai
                 for node in nodes])
        else: # assume an array of indices
            ai = self._node_info['_ai'][nodes]

        # the order they will be processed
        io_order = np.argsort(ai)
        ai = ai[io_order]
        # array to return them to original order
        return_order = np.empty_like(io_order)
        return_order[io_order] = np.arange(io_order.size)

        dfi = np.digitize(ai, self._node_io._ei)
        udfi = np.unique(dfi)
        data_files = [self.data_files[i] for i in udfi]
        index_list = [io_order[dfi == i] for i in udfi]

        return data_files, index_list, return_order

    def _node_io_loop_start(self, data_file):
        data_file._field_cache = {}
        data_file.open()

    def _node_io_loop_finish(self, data_file):
        data_file._field_cache = {}
        data_file.close()

    def _parse_parameter_file(self):
        self._prefix = \
          self.filename[:self.filename.rfind(self._suffix)]

        fh = h5py.File(self.filename, mode="r")
        for attr in ["hubble_constant",
                     "omega_matter",
                     "omega_lambda"]:
            setattr(self, attr, fh.attrs.get(attr, None))
        if "unit_registry_json" in fh.attrs:
            self.unit_registry = \
              UnitRegistry.from_json(
                  parse_h5_attr(fh, "unit_registry_json"))
        if "box_size" in fh.attrs:
            self.box_size = _hdf5_yt_attr(
                fh, "box_size", unit_registry=self.unit_registry)
        self.field_info.update(
            json.loads(parse_h5_attr(fh, "field_info")))
        self._size = fh.attrs["total_trees"]
        fh.close()

        # analysis fields in sidecar files
        analysis_filename = f"{self._prefix}-analysis{self._suffix}"
        if os.path.exists(analysis_filename):
            self.analysis_field_list = []
            self.analysis_filename = analysis_filename
            fh = h5py.File(analysis_filename, mode="r")
            analysis_fi = json.loads(parse_h5_attr(fh, "field_info"))
            fh.close()
            for field in analysis_fi:
                analysis_fi[field]["type"] = "analysis_saved"
                self.analysis_field_list.append(field)
            self.field_info.update(analysis_fi)
        else:
            self.analysis_filename = None

        self.field_list = list(self.field_info.keys())

    def _plant_trees(self):
        if self.is_planted:
            return

        fh = h5py.File(self.filename, "r")
        self._node_info['uid'][:] = fh["data"]["uid"][()].astype(np.int64)
        self._node_io._si = fh["index"]["tree_start_index"][()]
        self._node_io._ei = fh["index"]["tree_end_index"][()]
        fh.close()

        self._node_info['_ai'][:] = np.arange(self.size)
        self.data_files = \
          [YTreeDataFile(f"{self._prefix}_{i:04d}{self._suffix}")
           for i in range(self._node_io._si.size)]
        if self.analysis_filename is not None:
            for i, df in enumerate(self.data_files):
                df.analysis_filename = \
                  f"{self._prefix}_{i:04d}-analysis{self._suffix}"

    def get_yt_selection(self, above=None, below=None, equal=None, about=None,
                         conditionals=None, data_source=None):
        """
        Get a selection of halos meeting given criteria.

        This function can be used to create database-like queries to search
        for halos meeting various criteria. It will return a
        :class:`~yt.data_objects.selection_objects.cut_region.YTCutRegion`
        that can be queried to get field values for all halos meeting the
        selection criteria. The
        :class:`~yt.data_objects.selection_objects.cut_region.YTCutRegion`
        can then be passed to
        :func:`~ytree.frontends.ytree.arbor.YTreeArbor.get_nodes_from_selection`
        to get all the
        :class:`~ytree.data_structures.tree_node.TreeNode` objects that meet the
        criteria.

        If multiple criteria are provided, selected halos must meet all
        criteria.

        Parameters
        ----------
        above : optional, list of tuples with (field, value, <units>)
            Halos meeting a given criterion must have field values at or
            above the provided limiting value. Each entry in the list must
            contain the field name, limiting value, and (optionally) units.
        below : optional, list of tuples with (field, value, <units>)
            Halos meeting a given criterion must have field values at or
            below the provided limiting value. Each entry in the list must
            contain the field name, limiting value, and (optionally) units.
        equal : optional, list of tuples with (field, value, <units>)
            Halos meeting a given criterion must have field values equal to
            the provided value. Each entry in the list must contain the
            field name, value, and (optionally) units.
        about : optional, list of tuples with (field, value, tolerance, <units>)
            Halos meeting a given criterion must have field values within
            the tolerance of the provided value. Each entry in the list must
            contain the field name, value, tolerance, and (optionally) units.
        conditionals : optional, list of strings
            A list of conditionals for constructing a custom
            :class:`~yt.data_objects.selection_objects.cut_region.YTCutRegion`.
            This can be used instead of above/below/equal/about to create
            more complex selection criteria. See the Cut Regions section in the
            yt documentation for more information. The conditionals keyword
            can only be used if none of the first for selection keywords are
            given.
        data_source : optional, :class:`~yt.data_objects.data_containers.YTDataContainer`
            The source yt data container to be used to make the cut region.
            If none given, the
            :class:`~yt.data_objects.static_output.Dataset.all_data` container
            (i.e., the full dataset) is used. The ``ytds`` attribute is a yt
            dataset associated with the arbor and can be used to create data
            containers for use here.

        Returns
        -------
        cr : :class:`~yt.data_objects.selection_objects.cut_region.YTCutRegion`
            The cut region associated with the provided selection criteria.

        Examples
        --------
        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # select halos above 1e12 Msun at redshift > 0.5
        >>> sel = a.get_yt_selection(
        ...     above=[("mass", 1e13, "Msun"),
        ...            ("redshift", 0.5)])
        >>> print (sel["halos", "mass"])
        >>> print (sel["halos", "virial_radius"])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> ds = a.ytds
        >>> sphere = ds.sphere(ds.domain_center, (0.1, "unitary"))
        >>> # select halos below 1e13 Msun at redshift > 1
        >>> sel = a.get_yt_selection(
        ...     below=[("mass", 1e13, "Msun")],
        ...     above=[("redshift", 1)],
        ...     data_source=sphere)
        >>> print (sel["halos", "mass"])
        >>> print (sel["halos", "virial_radius"])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # select phantom halos (a consistent-trees field)
        >>> sel = a.get_yt_selection(equal=[("phantom", 1)])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # select halos with vmax of 200 +-10 km/s (i.e., 5%)
        >>> sel = a.get_yt_selection(about=[("vmax", 200, "km/s", 0.05)])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # use a yt conditional
        >>> sel = a.get_yt_selection(
        ...     conditionals=['obj["halos", "mass"] > 1e12'])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # select halos only within a sphere
        >>> ds = a.ytds
        >>> sphere = ds.sphere(ds.domain_center, (10, Mpc))
        >>> sel = a.get_yt_selection(
        ...     above=[("mass", 1e13)],
        ...     data_source=sphere)
        >>> # get the TreeNodes for the selection
        >>> for node in a.get_nodes_from_selection(sel):
        ...     print (node["mass"])

        See Also
        --------
        select_halos, get_nodes_from_selection

        """

        if above is None:
            above = []
        if below is None:
            below = []
        if equal is None:
            equal = []
        if about is None:
            about = []
        if conditionals is None:
            conditionals = []

        if not (bool(conditionals) ^ any([above, below, equal, about])):
            raise ValueError(
                "Must specify either conditionals or above/below/equal/about, not both."
                f"\nconditionals: {conditionals}"
                f"\nabove: {above}"
                f"\nbelow: {below}"
                f"\nequal: {equal}"
                f"\nabout: {about}")

        if data_source is None:
            data_source = self.ytds.all_data()

        if not isinstance(data_source, YTDataContainer):
            raise ValueError(
                f"data_source must be a YTDataContainer: {data_source}.")

        for criterion in above:
            condition = get_conditional("above", criterion)
            conditionals.append(condition)

        for criterion in below:
            condition = get_conditional("below", criterion)
            conditionals.append(condition)

        for criterion in equal:
            condition = get_conditional("equal", criterion)
            conditionals.append(condition)

        for criterion in about:
            conditions = get_about(criterion)
            conditionals.extend(conditions)

        cr = data_source.cut_region(conditionals)
        return cr

    def get_nodes_from_selection(self, container):
        """
        Generate TreeNodes from a yt data container.

        All halos contained within the data container will be
        returned as TreeNode objects. This returns a generator
        that can be iterated over or cast as a list.

        Parameters
        ----------
        container : :class:`~yt.data_objects.data_containers.YTDataContainer`
            Data container, such as a sphere or region, from
            which nodes will be generated.

        Returns
        -------
        nodes : generator
            The :class:`~ytree.data_structures.tree_node.TreeNode` objects
            contained within the container.

        Examples
        --------
        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> c = a.arr([0.5, 0.5, 0.5], "unitary")
        >>> sphere = a.ytds.sphere(c, (0.1, "unitary"))
        >>> for node in a.get_nodes_from_selection(sphere):
        ...     print (node["mass"])

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # select halos above 1e12 Msun at redshift > 0.5
        >>> sel = a.get_yt_selection(
        ...     above=[("mass", 1e13, "Msun"),
        ...            ("redshift", 0.5)])
        >>> my_nodes = list(a.get_nodes_from_selection(sel))

        """

        # Planting trees is necessary to get access to the start index for
        # each data file (i.e., self._node_io._si).
        self._plant_trees()

        all_storage = {}
        for my_store, my_chunk in parallel_objects(container.chunks([], "io"),
                                                   storage=all_storage):
            file_number = my_chunk['halos', 'file_number'].d.astype(int)
            file_root_index = my_chunk['halos', 'file_root_index'].d.astype(int)
            tree_index = my_chunk['halos', 'tree_index'].d.astype(int)
            arbor_index = self._node_io._si[file_number] + file_root_index
            my_store.result = (arbor_index, tree_index)

        # Gather results from all chunks.
        # This is quite fast so we can afford to pull everything back together.
        all_ai = []
        all_ti = []
        for my_i, my_values in sorted(all_storage.items()):
            all_ai.append(my_values[0])
            all_ti.append(my_values[1])
        all_ai = np.concatenate(all_ai).astype(int)
        all_ti = np.concatenate(all_ti).astype(int)
        all_results = list(zip(all_ai, all_ti))

        # Yield results as TreeNode objects.
        # The slowest part of this is generating the root nodes
        # with self._generate_root_node. The get_node call is fast.
        # We speed this up by creating a dictionary of roots so we
        # don't have to regenerate.
        # For an instant in time, this function accepted a trees kwarg
        # allowing the user to create and supply the list of all trees
        # ahead of time. However, I removed this because it was fairly
        # awkward and not much faster than the internal dict.

        # Note to the future, yielding just the (ai, ti) tuple as the minimum
        # required to generate a node later is basically instantaneous.
        # I experimented with yielding just (ai, ti) and then piping that
        # into parallel_trees. This was a speedup, but did not result in a
        # clean interface, so I have scrapped it from here for now.
        # In the future, it may be better to return a NodeContainer that
        # contains the (ai, ti) pairs and has parallel iteration implemented.

        my_trees = {}
        for ai, ti in all_results:
            if ai not in my_trees:
                root_node = self._generate_root_node(ai)
                my_trees[ai] = root_node
            root_node = my_trees[ai]
            yield root_node.get_node("forest", ti)

    def reload_arbor(self):
        """
        Reload the arbor and setup pre-existing derived fields.

        This can be useful to do after performing a long series of analyses
        in which many :ref:`analysis-fields` have been modified and the results
        saved in update mode (i.e., new results saved to an existing arbor).
        This will free analysis fields from memory as they now live on disk as
        conventional fields.

        This will automatically add any derived or analysis fields that were
        present in the original arbor. This function may prove useful for
        keeping memory usage under control in lengthy jobs. Note, this is only
        available for :ref:`ytree <load-ytree>` arbor types.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>> # do stuff
        >>> a = a.reload_arbor()
        """

        new_arbor = ytree_load(self.filename)

        add_fields = set(self.derived_field_list).difference(
            new_arbor.derived_field_list)
        for field in add_fields:
            fi = self.field_info[field].copy()
            name = fi.pop("name")
            function = fi.pop("function")
            del fi["type"], fi["dependencies"]
            new_arbor.add_derived_field(name, function, **fi)

        return new_arbor

    _ytds = None
    @property
    def ytds(self):
        """
        Load as a yt dataset.

        Merger tree data is loaded as a yt dataset, providing full access
        to yt functionality. Fields are accessed with the naming convention,
        ("halos", <field name>).

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("arbor/arbor.h5")
        >>>
        >>> ds = a.ytds
        >>> sphere = ds.sphere(ds.domain_center, (5, "Mpc"))
        >>> print (sphere["halos", "mass"])
        >>>
        >>> for node in a.get_nodes_from_selection(sphere):
        ...     print (node["position"])

        """
        if self._ytds is not None:
            return self._ytds
        with log_level(40, mylog=ytLogger):
            self._ytds = YTreeDataset(self.filename)
        return self._ytds

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .h5, be loadable as an hdf5 file,
        and have "arbor_type" attribute.
        """
        fn = args[0]
        if not fn.endswith(self._suffix):
            return False
        try:
            with h5py.File(fn, "r") as f:
                if "arbor_type" not in f.attrs:
                    return False
                atype = f.attrs["arbor_type"]
                if hasattr(atype, "astype"):
                    atype = atype.astype(str)
                if atype != "YTreeArbor":
                    return False
        except BaseException:
            return False
        return True



"""
TreeFarmArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np

from ytree.data_structures.io import \
    CatalogDataFile

class TreeFarmDataFile(CatalogDataFile):
    def open(self):
        self.fh = h5py.File(self.filename, "r")

    def _parse_header(self):
        self.open()
        fh = self.fh
        self.redshift = fh.attrs["current_redshift"]
        self.nhalos   = fh.attrs["num_halos"]
        # Files with no halos won't have the units.
        # Keep trying until we get one.
        if not hasattr(self.arbor, "field_list"):
            self._setup_field_info(fh)
        self.close()

    def _setup_field_info(self, fh):
        fields = list(fh.keys())
        fi = {}
        for field in fields:
            if fh[field].size == 0:
                # Zero-sized arrays won't have units, so don't bother.
                return
            units = fh[field].attrs["units"]
            if isinstance(units, bytes):
                units = units.decode("utf")
            fi[field] = {"source": "file", "units": units}

        fields.append("redshift")
        fi["redshift"] = {"source": "header", "units": ""}

        self.arbor.field_list = fields
        self.arbor.field_info.update(fi)

    def _read_data_default(self, rfields, dtypes):
        field_data = {}
        if not rfields:
            return field_data

        self.open()
        fh = self.fh
        field_data = dict((field, fh[field][()])
                          for field in rfields)
        self.close()

        for field in rfields:
            dtype = dtypes[field]
            field_data[field] = field_data[field].astype(dtype)
        return field_data

    def _read_data_select(self, rfields, tree_nodes, dtypes):
        field_data = {}
        if not rfields:
            return field_data

        file_ids = np.array([node._fi for node in tree_nodes])
        self.open()
        fh = self.fh
        for field in rfields:
            field_data[field] = fh[field][()][file_ids]
        self.close()

        for field in rfields:
            dtype = dtypes[field]
            field_data[field] = field_data[field].astype(dtype)
        return field_data



"""
tests for parallel_tree_nodes iterator



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import os

from ytree.utilities.testing import ParallelTest, TempDirTest

script_path = os.path.dirname(__file__)

class ParallelTreeNodesTest(TempDirTest, ParallelTest):
    test_script = os.path.join(script_path, "parallel/parallel_tree_nodes.py")
    arg_sets = (
        ("all", "forest", 0, 0),
        ("all", "tree",   0, 0),
        ("all", "prog",   0, 0),
        ("all", "forest", 2, 0),
        ("all", "forest", 0, 1),
        ("all", "nodes-10", 0, 0),
    )



"""
LHaloTree utilities



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
import os
import glob

from ytree.utilities.logger import \
    ytreeLogger

"""Default header data type."""
dtype_header_default = [
    # merger tree pointers
    ('Descendant', 'i4'),
    ('FirstProgenitor', 'i4'),
    ('NextProgenitor', 'i4'),
    ('FirstHaloInFOFgroup', 'i4'),
    ('NextHaloInFOFgroup', 'i4'),
    # properties of halo
    ('Len', 'i4'),
    ('M_Mean200', 'f4'),
    ('Mvir', 'f4'), # for Millennium, Mvir=M_Crit200
    ('M_TopHat', 'f4'),
    ('Pos', 'f4', 3),
    ('Vel', 'f4', 3),
    ('VelDisp', 'f4'),
    ('Vmax', 'f4'),
    ('Spin', 'f4', 3),
    ('MostBoundID', 'i8'),
    # original position in simulation tree files
    ('SnapNum', 'i4'),
    ('FileNr', 'i4'),
    ('SubhaloIndex', 'i4'),
    ('SubHalfMass', 'f4')]
halo_fields = ['Descendant', 'FirstProgenitor', 'NextProgenitor',
               'FirstHaloInFOFgroup', 'NextHaloInFOFgroup']


def read_header_default(filename):
    r"""Reads the default LHaloTree file header.

    Args:
        filename (str): Full path to file that header should be read from.

    Returns:
         tuple (int, np.ndarray): The size of the file header and the number
             of halos in each header.

    """
    # Open
    if isinstance(filename, str):
        fd = open(filename, 'rb')
        close = True
    else:
        fd = filename
        close = False
    # Read
    dtype1 = np.dtype([('ntrees', 'i4'), ('totnhalos', 'i4')])
    x1 = np.fromfile(fd, dtype=dtype1, count=1)
    ntrees = x1['ntrees'][0]
    nhalos = x1['totnhalos'][0]
    dtype2 = np.dtype('i4')
    x2 = np.fromfile(fd, dtype=dtype2, count=ntrees)
    assert (len(x2) == ntrees)
    assert (np.sum(x2) == nhalos)
    header_size = dtype1.itemsize + ntrees*dtype2.itemsize
    # Close
    if close:
        fd.close()
    return header_size, x2


def save_header_default(filename, nhalos_per_tree):
    r"""Writes the default LHaloTree file header.

    Args:
        filename (str): Full path to file that should be written to.
        nhalos_per_tree (np.ndarray): The number of halos in each tree that
            will be written to the file.
    
    Returns:
        int: The size of the header that was written.

    """
    ntrees = len(nhalos_per_tree)
    nhalos = np.sum(nhalos_per_tree)
    dtype1 = np.dtype([('ntrees', 'i4'), ('totnhalos', 'i4')])
    x1 = np.array([(ntrees, nhalos)], dtype=dtype1)
    x2 = nhalos_per_tree.astype('i4')
    header_size = x1.nbytes + x2.nbytes
    # Open
    if isinstance(filename, str):
        fd = open(filename, 'wb')
        close = True
    else:
        fd = filename
        close = False
    # Write
    x1.tofile(fd)
    x2.tofile(fd)
    # Close
    if close:
        fd.close()
    return header_size


def _read_from_mmap(filename, start=0, nhalo=None,
                    header_size=None, item_dtype=None):
    r"""Read one or more halos from a memmapped file.

    Args:
        filename (str, np.memmap): Either the full path to the file that should
            be read via memmap or an existing memmapped file.
        start (int, optional): Index of halo in memmap where read should start.
            Defaults to 0.
        nhalo (int, optional): Number of halos that should be read. Defaults to
            None and all halos following start are read.
        header_size (int, optional): Size of the header preceeding trees in the
            file. If this is not provided and filename is a path, it will be set
            to 0.
        item_dtype (np.dtype, optional): Data type of each halo entry. Defaults
            to dtype_header_default.

    Returns:
        np.ndarray: Structure array of halo data.

    """
    if isinstance(filename, np.memmap):
        mmap = filename
    else:
        if header_size is None:
            header_size = 0
        if item_dtype is None:
            item_dtype = dtype_header_default
        mmap = np.memmap(filename, dtype=item_dtype, mode='c', offset=header_size)
    if nhalo is None:
        stop = None
    else:
        stop = start + nhalo
    idx = slice(start, stop)
    return mmap[idx]


def _save_to_mmap(filename, data, start=0, header_size=None):
    r"""Save one or more halos to a memmapped file.

    Args:
        filename (str, np.memmap): Either the full path to the file that should
            be saved to via memmap or an existing memmapped file.
        data (np.ndarray): Structured array of halo data that should be saved.
        start (int, optional): Index in memmap where halos should be written.
            Defaults to 0.
        header_size (int, optional): Size of the header preceeding trees in the
            file. If this is not provided and filename is a path, it will be set
            to 0.

    """
    nhalo = len(data)
    if isinstance(filename, np.memmap):
        mmap = filename
        flush = False
    else:
        if header_size is None:
            header_size = 0
        item_dtype = data.dtype
        mmap = np.memmap(filename, dtype=item_dtype, mode='r+',
                         offset=header_size, shape=(nhalo, ))
        flush = True
    mmap[start:(start + nhalo)] = data[:]
    if flush:
        del mmap  # flush to disk


def _read_from_file(filename, start=0, nhalo=None,
                    header_size=None, item_dtype=None):
    r"""Read one or more halos from the file using np.fromfile.

    Args:
        filename (str, file): Either the full path to the file that should be
           openned and read or an existing open file object.
        start (int, optional): Index of halo in file (relative to the end of the
            header) where read should start. Defaults to 0.
        nhalo (int, optional): Number of halos that should be read. Defaults to
            all halos.
        header_size (int, optional): Size of the header preceeding trees in the
            file. Defaults to 0.
        item_dtype (np.dtype, optional): Data type of each halo entry. Defaults
            to dtype_header_default.

    Returns:
        np.ndarray: Structure array of halo data.

    """
    if header_size is None:
        header_size = 0
    if item_dtype is None:
        item_dtype = dtype_header_default
    item_dtype = np.dtype(item_dtype)
    # Open file as necessary
    opened = False
    if isinstance(filename, str):
        fd = open(filename, 'rb')
        opened = True
    else:
        fd = filename
    # Seek to halo location and read
    offset = header_size + (start * item_dtype.itemsize)
    fd.seek(offset, os.SEEK_SET)
    if nhalo is None:
        nhalo = -1
    out = np.fromfile(fd, dtype=item_dtype, count=nhalo)
    if opened:
        fd.close()
    return out


def _save_to_file(filename, data, start=0, header_size=None):
    r"""Save one or more halos to a file using np.tofile.

    Args:
        filename (str, file): Either the full path to the file that should be
           openned and saved to or an existing open file object.
        data (np.ndarray): Structured array of halo data that should be saved.
        start (int, optional): Index in file (in halso) where halos should be
            written. Defaults to 0.
        header_size (int, optional): Size of the header preceeding trees in the
            file. Defaults to 0.

    """
    if header_size is None:
        header_size = 0
    item_dtype = data.dtype
    # Open file as necessary
    opened = False
    if isinstance(filename, str):
        fd = open(filename, 'rb+')
        opened = True
    else:
        fd = filename
    # Seek to halo location and write
    offset = header_size + (start * item_dtype.itemsize)
    fd.seek(offset, os.SEEK_SET)
    data.tofile(fd)
    if opened:
        fd.close()


def read_trees_default(filename, **kwargs):
    r"""Read trees from a file.

    Args:
        filename (str, np.memmap): Either the full path to the file that should
            be read via memmap or an existing memmapped file.
        **kwargs: Additional keyword arguments are passed to _read_from_mmap.

    Returns:
        dict: Arrays of fields for each halo.

    """
    # out_ra = _read_from_file(filename, **kwargs)
    out_ra = _read_from_mmap(filename, **kwargs)
    # if len(out_ra) == 1:
    #     for k in out_ra.dtype.fields.keys():
    #         if not isinstance(out_ra[k], np.ndarray):
    #             print(k, out_ra[k])
    #             raise Exception
    #     out = {k: out_ra[k].copy() for k in out_ra.dtype.fields.keys()}
    # else:
    out = {k: out_ra[k].copy() for k in out_ra.dtype.fields.keys()}
    return out


def save_trees_default(filename, data, item_dtype=None, **kwargs):
    r"""Save trees to a file.

    Args:
        filename (str, np.memmap): Either the full path to the file that should
            be saved to via memmap or an existing memmapped file.
        data (dict): Arrays of fields for each halo.
        item_dtype (np.dtype, optional): Data type of each halo entry. Defaults
            to dtype_header_default.
        **kwargs: Additional keyword arguments are passed to _save_to_mmap.

    """
    if item_dtype is None:
        item_dtype = dtype_header_default
    item_dtype = np.dtype(item_dtype)
    nhalo = len(data[next(iter(data))])  # Gross, but Py 2 & 3 compat
    data_ra = np.empty(nhalo, dtype=item_dtype)
    for k in item_dtype.fields.keys():
        data_ra[k] = data[k]
    # return _save_to_file(filename, data_ra, **kwargs)
    return _save_to_mmap(filename, data_ra, **kwargs)


class LHaloTreeReader:
    r"""Class for reading halos from an LHaloTree file.

    Args:
        filename (str): Full path to file that trees should be read from.
        parameters (dict, optional): Dictionary of simulation parameters. If not
            provided, parameters are loaded from parameter_file.
        parameter_file (str, optional): Full path to file that contains
            simulation parameters. If not provided, one with the suffix '.param'
            is searched for in the same directory as the tree file. An IOError
            will be raised if it cannot be found.
        scale_factors (list or np.ndarray, optional): 1D array of scale factors for each
            snapshot. If not provided, they are loaded from scale_factor_file.
        scale_factor_file (str, optional): Full path to the file containing the
            list of scale factors for each snapshot. If not provided, one with
            the suffix '.a_list' is searched for in the same directory as the
            tree file. An IOError will be raised if it cannot be found.
        header_size (int, optional): The size of the file header. If not
            provided, it will be obtained from read_header_func.
        nhalos_per_tree (np.ndarray, optional): The number of halos in each of
            the ntree trees in the file. If not provided, it will be obtained
            from read_header_func.
        read_header_func (obj, optional): Callable object that takes a filename
            and returns the size of the file header and the number of halos in
            each tree in the file. Defaults to read_header_default.
        item_dtype (obj, optional): np.dtype or object that can be converted
            into a numpy dtype specifying the structure of single halo entries
            in the tree. Defaults to dtype_header_default.
        silent (bool, optional): If True, diagnositic info is not printed.
            Defaults to False.
        validate (bool, optional): If True, the data in the file will be
            validated and errors will be raised if there are inconsistencies.
            Defaults to False.

    Attributes:
        filename (str): Full path to file that trees should be read from.
        fileindex (int): Index of this file in the full set of trees.
        filepattern (str): Glob style pattern that can be used to find other
            files in the full set of trees.
        parameter_file (str): Full path to file that contains simulation
            parameters.
        header_size (int): The size of the file header.
        nhalos_per_tree (np.ndarray): The number of halos in each of the ntree
            trees in the file.
        nhalos_before_tree (np.ndarray): The number of halos before each of the
            ntree trees in the file.
        totnhalos (int): Total number of halos in this file.
        ntrees (int): Number of trees in this file.
        item_dtype (np.dtype): Data type specifying the structure of single halo
            entries in the tree.
        raw_fields (list): Available fields that are present for each halo based
            on item_dtype.
        add_fields (list): Calculated fields that will be added to trees as they
            are read.
        fields (list): All available fields that are present for each halo.
        filenum (int): Index of this file in the complete set for all trees.
        scale_factor_file (str): Full path to the file containing the list of
            scale factors for each snapshot.

    Raises:
        IOError: If filename is not a valid path.
        IOError: If a parameter_file is not provided or is not a valid path.
        IOError: If a scale_factor_file is not provided and cannot be found.
        IOError: If the file dosn't have the expected number of bytes.

    """

    def __init__(self, filename, parameters=None, parameter_file=None,
                 scale_factors=None, scale_factor_file=None,
                 header_size=None, nhalos_per_tree=None, read_header_func=None,
                 item_dtype=None, silent=False, validate=False):
        # Files
        self.filename = self._verify_file(filename)
        self.fileindex = 0
        self.filepattern = self.filename
        ext = self.filename.split('.')[-1]
        if ext.isdigit():
            self.fileindex = int(ext)
            self.filepattern = os.path.splitext(self.filename)[0] + ".*"
        if parameters is not None:
            self._parameters = parameters
            self.parameter_file = None
        else:
            self.parameter_file = self._verify_file(
                parameter_file, suffix='.param', error_tag='Parameter file',
                silent=silent)
            self._parameters = None
        if scale_factors is not None:
            self._scale_factors = np.asarray(scale_factors)
            self.scale_factor_file = None
        else:
            self.scale_factor_file = self._verify_file(
                scale_factor_file, suffix='.a_list', error_tag='Scale factor file',
                silent=silent)
            self._scale_factors = None
        # Header info
        if (header_size is None) or (nhalos_per_tree is None):
            if (read_header_func is None):
                read_header_func = read_header_default
            header_size, nhalos_per_tree = read_header_func(filename)
        if (item_dtype is None):
            item_dtype = dtype_header_default
        self.header_size = header_size
        self.nhalos_per_tree = nhalos_per_tree
        self.nhalos_before_tree = (np.cumsum(self.nhalos_per_tree) -
                                   self.nhalos_per_tree)
        self.totnhalos = np.sum(self.nhalos_per_tree)
        self.ntrees = len(self.nhalos_per_tree)
        self.item_dtype = np.dtype(item_dtype)
        # Fields
        self.raw_fields = list(self.item_dtype.fields.keys())
        self.add_fields = ['desc_uid', 'scale_factor', 'uid',
                           'x', 'y', 'z', 'vx', 'vy', 'vz',
                           'Jx', 'Jy', 'Jz']
        self.fields = self.raw_fields + self.add_fields
        for k in ['Pos', 'Vel', 'Spin']:
            self.fields.remove(k)
        # Check file size
        item_size = self.item_dtype.itemsize
        body_size = self.totnhalos * item_size
        file_size = os.stat(self.filename).st_size
        if body_size != (file_size - self.header_size):  # pragma: no cover
            raise IOError(
                f"File is {file_size} bytes, but {self.totnhalos} items of size "
                f"{item_size} with header of {self.header_size} bytes should be "
                f"{body_size + self.header_size} bytes total.")
        # Load all data, validate, and cache some fields
        self.set_global_properties(validate=validate)

    def set_global_properties(self, validate=False):
        r"""Set attributes for all trees by loading all of the halos.

        Args:
            validate (bool, optional): If True, the data loaded from the
                file will be validated. Defaults to False.

        .. todo:: For small files, the data could be cached which would
                  greatly speed up loading.

        """
        # Tree num array
        self.treenum_arr = np.zeros(self.totnhalos, dtype='int64')
        start = self.nhalos_before_tree
        stop = start + self.nhalos_per_tree
        for t in range(self.ntrees):
            self.treenum_arr[start[t]:stop[t]] = t
        # Memmap/file object
        self.fobj = np.memmap(self.filename, dtype=self.item_dtype, mode='c',
                              offset=self.header_size)
        # Read all data
        data = self.read_all_trees(skip_add_fields=True, validate=validate)
        # File number
        self.filenum = data['FileNr'][0]
        if (data['SnapNum'][0] + 1) != len(self.scale_factors):  # pragma: no cover
            ytreeLogger.warning(
                f"First FoF central is in snapshot {data['SnapNum'][0] + 1}/{len(self.scale_factors)}.")
        # Halo unique IDs
        self.all_uids = np.bitwise_or(
            np.int64(self.filenum) << 32, np.arange(self.totnhalos, dtype='int64'))
        # Get descendant unique IDs
        desc = data['Descendant']
        pos_flag = (desc >= 0)
        desc_uid = np.zeros(self.totnhalos, dtype='int64') - 1
        desc_abs = self.get_total_index(self.treenum_arr, desc)
        desc_uid[pos_flag] = self.all_uids[desc_abs[pos_flag]]
        self.all_desc_uids = desc_uid
        # Add fields and cache root fields
        data = self.add_computed_fields(-1, data, validate=validate)
        root_idx = self.nhalos_before_tree
        self._root_data = dict()
        for k in self.fields:
            self._root_data[k] = data[k][root_idx]

    def _verify_file(self, filename, suffix=None, error_tag=None, silent=False):
        r"""Verify that the provided file exists. If it is None, and a suffix
        is provided, the file will be searched for in the same directory as
        self.filename.

        Args:
            filename (str): Full path that should be verified or None if the
                file should be searched for.
            suffix (str, optional): Suffix that should be used to search for a
                file if filename is None or dosn't exist. Defaults to None and
                no search is performed.
            error_tag (str, optional): Tag to use for error message in case
                where file dosn't exist. Defaults to None and 'File' is used.
            silent (bool, optional): If True, messages about located files are
                not printed. Defaults to False.

        Returns:
            str: Verified full path to the file.

        Raises:
            IOError: If filename is not a valid path and suffix is None.
            IOError: If a filename matching suffix cannot be located.

        """
        if error_tag is None:
            error_tag = 'File'
        if (filename is None) or (not os.path.isfile(filename)):
            if suffix is None:  # pragma: no cover
                raise IOError(f"{error_tag} dosn't exist: {filename}")
            pattern = os.path.join(os.path.dirname(self.filename), '*' + suffix)
            files = glob.glob(pattern)
            if len(files) == 0:  # pragma: no cover
                raise IOError(f"{error_tag} could not be located matching: {pattern}")
            else:
                filename = files[0]
                if not silent:
                    print(f"Using {error_tag.lower()} found at {filename}")
        return filename

    @property
    def parameters(self):
        r"""dict: Key/value pairs read from the parameter file."""
        if self._parameters is None:
            self._parameters = dict()
            with open(self.parameter_file, 'r') as fd:
                for line in fd:
                    line_strip = line.strip()
                    if len(line_strip) == 0:
                        continue
                    line_vars = line_strip.split()
                    if len(line_vars) != 2:
                        continue
                    self._parameters[line_vars[0]] = line_vars[1]
        return self._parameters

    @property
    def hubble_constant(self):
        r"""float: Hubble constants."""
        return float(self.parameters['HubbleParam'])

    @property
    def omega_matter(self):
        r"""float: Matter energy density."""
        return float(self.parameters['Omega0'])

    @property
    def omega_lambda(self):
        r"""float: Dark matter energy density."""
        return float(self.parameters['OmegaLambda'])

    @property
    def box_size(self):
        r"""float: Size of periodic boundaries."""
        return float(self.parameters['BoxSize'])

    @property
    def periodic(self):
        r"""bool: Does the simulation have periodic boundaries or not."""
        return bool(int(self.parameters['PeriodicBoundariesOn']))

    @property
    def comoving(self):
        r"""bool: Is the simulation in comoving units."""
        return bool(int(self.parameters['ComovingIntegrationOn']))

    @property
    def units_vel(self):
        r"""str: Units of velocity."""
        u_cms = float(self.parameters['UnitVelocity_in_cm_per_s'])
        out = f"{u_cms / 100000.0:f}*km/s"
        # TODO: Does this need adjusted for comoving?
        return out

    @property
    def units_len(self):
        r"""str: Units of length."""
        u_cm = float(self.parameters['UnitLength_in_cm'])
        out = f"{u_cm / 3.085678e+21:f}*kpc"
        if self.comoving:
            out += "/h"
        return out

    @property
    def units_mass(self):
        r"""str: Units of mass."""
        u_g = float(self.parameters['UnitMass_in_g'])
        out = f"{u_g / 1.989e+33:f}*Msun"
        if self.comoving:
            out += "/h"
        return out

    @property
    def scale_factors(self):
        r"""np.ndarray: Array of scale factors at each snapshot."""
        if self._scale_factors is None:
            self._scale_factors = np.fromfile(self.scale_factor_file, sep='\n')
        return self._scale_factors

    def get_total_index(self, treenum, halonum=None):
        r"""Get the slice that selects halos in a single tree.

        Args:
            treenum (int, np.ndarray): Index/indicies of tree(s) to get the
                total index for If -1, a slice that includes all halos in the
                file will be returned.
            halonum (int, np.ndarray, optional): Index/indices of halo(s) within
                the tree(s) to get the total index for in the file. Defaults to
                None and the slice for the entire tree is returned.

        Returns:
            slice, int: Slice/indices of halo(s) for this tree/halo(s).

        """
        if (not isinstance(treenum, np.ndarray)) and (treenum == -1):
            return slice(0, self.totnhalos)
        start = self.nhalos_before_tree[treenum]
        if halonum is None:
            return slice(start, start + self.nhalos_per_tree[treenum])
        else:
            return start + halonum

    # def get_tree_offset(self, treenum):
    #     r"""Get the offset in bytes of a tree from the beginning of the file.

    #     Args:
    #         treenum (int): Index of the tree to compute the offset of.

    #     Returns:
    #         int: Number of bytes from beginning of the file to the start of the
    #             requested tree.

    #     """
    #     if treenum == -1:
    #         return self.header_size
    #     offset = self.header_size + (self.nhalos_before_tree[treenum] *
    #                                  self.item_dtype.itemsize)
    #     return offset

    # def get_halo_offset(self, treenum, halonum):
    #     r"""Get the offset in bytes of a halo from the beginning of the file.

    #     Args:
    #         treenum (int): Index of the tree that requested halo is in.
    #         halonum (int): Index of the requested in halo in the tree.

    #     Returns:
    #         int: Number of bytes from beginning of the file to the start of the
    #             requested halo.

    #     """
    #     offset = self.get_tree_offset(treenum)
    #     if halonum is not None:
    #         offset += halonum * self.item_dtype.itemsize
    #     return offset

    def read_single_tree(self, treenum, halonum=None, fd=None,
                         skip_add_fields=False, validate=False):
        r"""Read a single tree from the file.

        Args:
            treenum (int): Index of the tree that should be returned. If -1,
                data for every tree in the file will be returned.
            halonum (int, optional): If provided, this is the index of a
                particular halo within the tree that should be returned. If not
                provided, the entire tree is returned.
            fd (file, optional): Open file identifier. If not provided, the file
                is opened for the read and then closed.
            skip_add_fields (bool, optional): If True, the calculated fields
                will not be added or checked for. Defaults to False.
            validate (bool, optional): If True, the resulting data will be
                validated. Defaults to False.

        Returns:
            dict: Dictionary of fields for each halo in the file/tree/halo.

        """
        if treenum >= 0:
            start = self.nhalos_before_tree[treenum]
            if halonum is None:
                nhalo = self.nhalos_per_tree[treenum]
            else:
                nhalo = 1
        else:
            start = 0
            nhalo = self.totnhalos
        # Read from map/file
        out = read_trees_default(self.fobj, start=start, nhalo=nhalo)
        # Validate
        if validate:
            self.validate_tree(treenum, out, halonum=halonum)
        # Add fields
        if not skip_add_fields:
            out = self.add_computed_fields(treenum, out, halonum=halonum,
                                           validate=validate)
        return out

    def read_single_halo(self, treenum, halonum, **kwargs):
        r"""Read a single halo entry from a tree in the file.

        Args:
            treenum (int): Index of the tree that contains the requested halo.
            halonum (int): Index of the halo within its tree.
            **kwargs: Additional keyword arguments are passed to
                read_single_tree.

        Returns:
            np.ndarray, dict: Dictionary or single element structured array with
                fields for the corresponding halo.

        """
        return self.read_single_tree(treenum, halonum=halonum, **kwargs)

    def read_single_root(self, treenum, **kwargs):
        r"""Read an entry for a single tree root. First check to see if the root
        was chached.

        Args:
            treenum (int): Index of the tree that data should be returned for.
                If -1, the data for every root in the file is returned.
            **kwargs: Additional keyword arguments are passed to
                read_single_tree in the event that the root data is not cached.

        Returns:
            np.ndarray, dict: Dictionary or single element structured array with
                fields for the corresponding root.

        """
        root_data = self._root_data
        if treenum == -1:
            return root_data
        out = dict()
        for k, v in root_data.items():
            out[k] = np.array([v[treenum]], dtype=v.dtype)
        return out

    def read_all_trees(self, **kwargs):
        r"""Read a all lhalotrees from the file.

        Args:
            **kwargs: All keyword arguments are passed to read_single_tree.

        Returns:
            dict, np.ndarray: Dictionary or structured array with fields for
                each halo in the file.

        """
        return self.read_single_tree(-1, halonum=None, **kwargs)

    def add_computed_fields(self, treenum, tree, halonum=None, validate=False):
        r"""Add computed fields to the numpy array.

        Args:
            treenum (int): Index of tree in file.
            tree (dict): Dictionary of fields for each halo in the tree.
            halonum (int, optional): Index of halo in the tree if tree contains
                a single halo.
            validate (bool, optional): If True, the resulting data will be
                validated. Defaults to False.

        Returns:
            dict: Dictionary of fields for each halo with added fields.

        """
        nhalos = len(tree['SnapNum'])
        if treenum == -1:
            assert (nhalos == self.totnhalos)
            assert (halonum is None)
        elif nhalos != self.nhalos_per_tree[treenum]:
            assert (nhalos == 1)
            assert (halonum is not None)
        else:
            halonum = None
        # Unique ID for each halo and descendant in tree
        idx = self.get_total_index(treenum, halonum)
        uid = self.all_uids[idx]
        desc_uid = self.all_desc_uids[idx]
        # Scale factors
        scale_factor = self.scale_factors[tree['SnapNum']]
        # Position x, y, z
        x = tree['Pos'][:, 0]
        y = tree['Pos'][:, 1]
        z = tree['Pos'][:, 2]
        # Velocity x, y, z
        vx = tree['Vel'][:, 0]
        vy = tree['Vel'][:, 1]
        vz = tree['Vel'][:, 2]
        # Spin x, y, z
        Jx = tree['Spin'][:, 0]
        Jy = tree['Spin'][:, 1]
        Jz = tree['Spin'][:, 2]
        # Add new fields
        new_fields = dict(uid=uid, desc_uid=desc_uid, scale_factor=scale_factor,
                          x=x, y=y, z=z, vx=vx, vy=vy, vz=vz, Jx=Jx, Jy=Jy, Jz=Jz)
        out = tree
        out.update(**new_fields)
        if validate:
            self.validate_fields(out)
        return out

    def validate_tree(self, treenum, tree, halonum=None):
        r"""Check that the tree conforms to expectation.

        Args:
            treenum (int): Index of the tree being validated in the file.
            tree (dict): Dictionary of fields for each halo in the tree.
            halonum (int): Index of halo in tree if this is a single halo.
                Defaults to None and tree is treated as a complete tree.

        Raises:
            AssertionError: If the tree does not have the expected number of
                halos.
            AssertionError: If the tree does not have all of the expected fields
                from the default LHaloTree format.
            AssertionError: If any of the halos have fields pointing to halos
                outside the tree (indicies exceeding nhalos).
            AssertionError: If there are any halos in a FOF group that are not
                in the same snapshot as the central.
            AssertionError: If the progenitors/descendants of any halo are not
                backwards/forwards in time.

        """
        # Check size
        if treenum == -1:
            nhalos = self.totnhalos
            halonum = None
        else:
            if halonum is not None:
                nhalos = 1
            else:
                nhalos = self.nhalos_per_tree[treenum]
        fields = list(tree.keys())
        for k in fields:
            assert (len(tree[k]) == nhalos)
        # Check fields
        for k in self.raw_fields:
            assert (k in fields)
        # Don't check tree indices for a single halo
        if halonum is not None:
            return
        # For all trees get list of local nhalos for every halo
        if treenum == -1:
            treenum_arr = self.treenum_arr
            nhalos = self.nhalos_per_tree[treenum_arr]
        # Check that halos are within tree
        for k in halo_fields:
            assert ((tree[k] < nhalos).all())
        # Check FOF central exists and all subs in one snapshot
        central = tree['FirstHaloInFOFgroup']
        assert ((central >= 0).all())
        if treenum == -1:
            central = self.get_total_index(treenum_arr, central)
        assert ((tree['SnapNum'] == tree['SnapNum'][central]).all())
        # Check that progenitors/descendants are back/forward in time
        descend = tree['Descendant'].astype('i4')
        has_descend = (descend >= 0)
        not_descend = np.logical_not(has_descend)
        # Not strictly True
        # assert ((tree['SnapNum'][not_descend] ==
        #         (len(self.scale_factors) - 1)).all())
        if treenum == -1:
            descend = self.get_total_index(treenum_arr, descend)
        assert ((tree['SnapNum'][descend[has_descend]] >
                tree['SnapNum'][has_descend]).all())
        # Check progenitors are back in time
        descend[not_descend] = np.where(not_descend)[0]
        progen1 = tree['FirstProgenitor']
        progen2 = tree['NextProgenitor']
        has_progen1 = (progen1 >= 0)
        has_progen2 = (progen2 >= 0)
        if treenum == -1:
            progen1 = self.get_total_index(treenum_arr, progen1)
            progen2 = self.get_total_index(treenum_arr, progen2)
        assert ((tree['SnapNum'][progen1[has_progen1]] <=
                tree['SnapNum'][descend[has_progen1]]).all())
        assert ((tree['SnapNum'][progen2[has_progen2]] <=
                tree['SnapNum'][descend[has_progen2]]).all())

    def validate_fields(self, tree):
        r"""Check that the tree/halo has all of the expected fields.

        Args:
            tree (dict): Dictionary of fields for each halo in the tree.

        Raises:
            AssertionError: If the tree does not have all of the expected fields.

        """
        # Check fields
        fields = list(tree.keys())
        for k in self.fields:
            assert (k in fields)



"""
LHaloTreeHDF5Arbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import glob
import h5py
import numpy as np
import re

from yt.funcs import \
    get_pbar

from ytree.data_structures.arbor import \
    SegmentedArbor

from ytree.frontends.lhalotree_hdf5.fields import \
    LHaloTreeHDF5FieldInfo
from ytree.frontends.lhalotree_hdf5.io import \
    LHaloTreeHDF5DataFile, \
    LHaloTreeHDF5TreeFieldIO

class LHaloTreeHDF5Arbor(SegmentedArbor):
    """
    IllustrisTNG format.
    """

    _suffix = ".hdf5"
    _data_file_class = LHaloTreeHDF5DataFile
    _field_info_class = LHaloTreeHDF5FieldInfo
    _tree_field_io_class = LHaloTreeHDF5TreeFieldIO

    def __init__(self, filename,
                 hubble_constant=1.0, box_size=None,
                 omega_matter=None, omega_lambda=None):
        self.hubble_constant = hubble_constant
        self.omega_matter = omega_matter
        self.omega_lambda = omega_lambda
        if box_size is not None:
            self.box_size = self.quan(box_size, "Mpc/h")
        super().__init__(filename)

    def _get_data_files(self):
        suffix = self._suffix
        reg = re.search(rf"^(.+\D)\d+{suffix}$", self.filename)
        if reg is None:
            raise RuntimeError(
                f"Cannot determine numbering system for {self.filename}.")
        prefix = reg.groups()[0]
        files = glob.glob(f"{prefix}*{self._suffix}")
        self.data_files = [self._data_file_class(f, self) for f in files]
        self._file_count = np.array([df._size for df in self.data_files])
        self._size = self._file_count.sum()

    def _parse_parameter_file(self):
        f = h5py.File(self.parameter_filename, mode='r')
        g = f["Header"]
        ntrees = g.attrs["NtreesPerFile"]
        self._redshifts = g["Redshifts"][()]

        field_list = []
        if ntrees == 0:
            field_list = []
            return

        g = f["Tree0"]
        for d in g:
            dshape = g[d].shape
            if len(dshape) == 1:
                field_list.append(d)
            else:
                field_list.extend(
                    [f"{d}_{i}" for i in range(dshape[1])])

        self.field_list = field_list
        fi = dict((field, {}) for field in field_list)
        for field in ["uid", "desc_uid"]:
            self.field_list.append(field)
            fi[field] = {"units": "", "source": "arbor"}
        fi["desc_uid"]["dependencies"] = ["Descendant"]
        self.field_info.update(fi)

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        c = 0
        file_offsets = self._file_count.cumsum() - self._file_count
        pbar = get_pbar('Planting trees', self._size)
        for idf, data_file in enumerate(self.data_files):
            data_file.open()
            tree_size = data_file.fh["Header"]["TreeNHalos"][()]
            data_file.close()

            size = data_file._size
            istart = file_offsets[idf]
            iend = istart + size

            self._node_info['_fi'][istart:iend] = idf
            self._node_info['_si'][istart:iend] = np.arange(size)
            self._node_info['_tree_size'][istart:iend] = tree_size
            c += size
            pbar.update(c)
        pbar.finish()
        uids = self._node_info['_tree_size']
        self._node_info['uid'] = uids.cumsum() - uids

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fn = args[0]

        if not h5py.is_hdf5(fn):
            return False

        attrs = ["FirstSnapshotNr", "LastSnapshotNr", "SnapSkipFac",
                 "NtreesPerFile", "NhalosPerFile", "ParticleMass"]
        groups = ["Redshifts", "TotNsubhalos", "TreeNHalos"]

        with h5py.File(fn, mode='r') as f:
            g = f["Header"]
            for attr in attrs:
                if attr not in g.attrs:
                    return False
            for group in groups:
                if group not in g:
                    return False
        return True



"""
LHaloTreeArbor io classes and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
# import weakref
from ytree.data_structures.io import \
    FieldIO, \
    TreeFieldIO


# class LHaloTreeFileID:
#     r"""Handle cached file ID.

#     Args:
#         filename (str): Full path to file that should be opened for reading.

#     Attributes:
#         fd (file): Open file descriptor.

#     """
#     def __init__(self, filename):
#         self._finalizer = None
#         self.fd = None
#         self.update(filename)

#     def close(self):
#         r"""Close the file descriptor."""
#         self._finalizer()

#     def update(self, filename):
#         r"""Update the file descriptor and finalizer.

#         Args:
#             filename (str): Full path to file that should be opened for reading.

#         """
#         if (self.fd is None) or (self.fd.name != filename):
#             if self.fd is not None:
#                 self.close()
#             self.fd = open(filename, 'rb')
#             self._finalizer = weakref.finalize(self, self.fd.close)


class LHaloTreeTreeFieldIO(TreeFieldIO):

    def _read_fields(self, root_node, fields, dtypes=None,
                     f=None, root_only=False):
        """
        Read fields from disk for a single tree.

        Here we accept a list of fields and return a dictionary of NumPy
        arrays for each field.

        dtypes will be an optional dictionary of type for each field

        f will optionally be the already-opened file handle.

        If root_only is true, we only want the field value for the root
        of the tree.

        Below is the example for ctrees.
        """
        if dtypes is None:
            dtypes = {}

        lht = root_node._lht

        # This stores the file ID in a class that handles clean up via
        # weakref.finalize. The cached file object is checked to see if it
        # access the correct file for this node. If not, the cached file
        # object is closed and replaced with a file object for the correct
        # file.
        # This is only necessary if LHaloTreeReader is using np.fromfile (it
        # currently uses np.memmap).
        # Get cached fid.
        # if f is None:
        #     if not hasattr(self, '_cached_fid'):
        #         self._cached_fd = LHaloTreeFileID(lht.filename)
        #     self._cached_fd.update(lht.filename)
        #     f = self._cached_fd.fd

        # Don't read all data if only uid/desc_uid requested
        data = None
        for field in fields:
            if field not in ['uid', 'desc_uid']:
                # if root_only:
                #     data = lht.read_single_root(root_node._index_in_lht, fd=f)
                # else:
                data = lht.read_single_tree(root_node._index_in_lht, fd=f)
                break
        if data is None:
            data = dict()
            # always get all ids, even if root_only, since it's fast.
            halonum = None
            tot_idx = lht.get_total_index(root_node._index_in_lht, halonum)
            data['uid'] = lht.all_uids[tot_idx]
            data['desc_uid'] = lht.all_desc_uids[tot_idx]
        field_data = {field: data[field] for field in fields}

        self._apply_units(fields, field_data)

        return field_data


class LHaloTreeRootFieldIO(FieldIO):
    def _read_fields(self, storage_object, fields, dtypes=None):
        r"""Add root fields in bulk to same time."""
        if dtypes is None:
            dtypes = {}
        my_dtypes = self._determine_dtypes(fields, override_dict=dtypes)

        field_data = {field: np.empty(self.arbor.size, dtype=my_dtypes[field])
                      for field in fields}

        self._apply_units(fields, field_data)

        ntrees_prev = 0
        for lht in self.arbor._lhtfiles:
            ntrees = lht.ntrees
            for field in fields:
                field_data[field][ntrees_prev:(ntrees_prev + ntrees)] = \
                  lht._root_data[field]
            ntrees_prev += ntrees

        return field_data



"""
TreeNodeSelector functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.utilities.operator_registry import \
    OperatorRegistry

tree_node_selector_registry = OperatorRegistry()

def add_tree_node_selector(name, function):
    r"""
    Add a TreeNodeSelector to the registry of known selectors, so they
    can be chosen with :func:`~ytree.data_structures.arbor.Arbor.set_selector`.

    Parameters
    ----------
    name : string
        Name of the selector.
    function : callable
        The associated function.

    Examples
    --------

    >>> import ytree
    >>> def max_value(ancestors, field):
    ...     vals = np.array([a[field] for a in ancestors])
    ...     return ancestors[np.argmax(vals)]
    >>> ytree.add_tree_node_selector("max_field_value", max_value)
    >>> a = ytree.load("tree_0_0_0.dat")
    >>> a.set_selector("max_field_value", "mass")
    >>> print (a[0]["prog"])

    """
    tree_node_selector_registry[name] = TreeNodeSelector(function)

class TreeNodeSelector:
    r"""
    The TreeNodeSelector is responsible for choosing which one of a
    halo's ancestors to return when querying the line of main
    progenitors for a halo.

    Parameters
    ----------
    ancestors : list of TreeNode objects
        List of TreeNode objects from which to select.

    The function should return a single TreeNode.

    Examples
    --------

    >>> import ytree
    >>> def max_value(ancestors, field):
    ...     vals = np.array([a[field] for a in ancestors])
    ...     return ancestors[np.argmax(vals)]
    >>> ytree.add_tree_node_selector("max_field_value", max_value)
    >>> a = ytree.load("tree_0_0_0.dat")
    >>> a.set_selector("max_field_value", "mass")
    >>> print (a[0]["prog"])

    """
    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None: self.args = []
        self.kwargs = kwargs
        if self.kwargs is None: self.kwargs = {}

    def __call__(self, ancestors):
        return self.function(ancestors, *self.args, **self.kwargs)

def max_field_value(ancestors, field):
    r"""
    Return the TreeNode with the maximum value of the given field.

    Parameters
    ----------
    ancestors : list of TreeNode objects
        List of TreeNode objects from which to select.
    field : string
        Field to be used for selection.

    Returns
    -------
    TreeNode object

    """

    vals = np.array([a[field] for a in ancestors])
    return ancestors[np.argmax(vals)]


add_tree_node_selector("max_field_value", max_field_value)

def min_field_value(ancestors, field):
    r"""
    Return the TreeNode with the minimum value of the given field.

    Parameters
    ----------
    ancestors : list of TreeNode objects
        List of TreeNode objects from which to select.
    field : string
        Field to be used for selection.

    Returns
    -------
    TreeNode object

    """

    vals = np.array([a[field] for a in ancestors])
    return ancestors[np.argmin(vals)]


add_tree_node_selector("min_field_value", min_field_value)



"""
save_arbor supporting functions



"""

import json
import numpy as np
import os
import types

from yt.frontends.ytdata.utilities import save_as_dataset
from yt.funcs import get_pbar
from ytree.utilities.io import ensure_dir
from ytree.utilities.logger import ytreeLogger as mylog

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

def save_arbor(arbor, filename=None, fields=None, trees=None,
               save_in_place=None, save_roots_only=False,
               max_file_size=524288):
    """
    Save the arbor to a file.

    This is the internal function called by Arbor.save_arbor.
    """

    from ytree.frontends.ytree.arbor import YTreeArbor
    is_ytree = isinstance(arbor, YTreeArbor)
    if save_in_place is None:
        save_in_place = is_ytree

    elif save_in_place and not is_ytree:
        raise ValueError(
            f"Cannot do save_in_place with arbor type {type(arbor)}. "
            "Resave full arbor first and then reload.")

    if not save_in_place and save_roots_only:
        raise ValueError(
            "Cannot have save_in_place=False with save_roots_only=True")

    if isinstance(trees, types.GeneratorType):
        trees = list(trees)

    arbor._plant_trees()
    update, filename = determine_save_state(
        arbor, filename, fields, trees, save_in_place)
    filename = determine_output_filename(filename, ".h5")
    fields = determine_field_list(arbor, fields, update)

    if not fields:
        mylog.warning(
            "No action will be taken for the following reasons:\n"
            " - This dataset is already a YTreeArbor.\n"
            " - No filename has been given.\n"
            " - No new analysis fields have been created.\n"
            " - No custom list of trees has been provided.")
        return None

    group_nnodes, group_ntrees, root_field_data = \
      save_data_files(arbor, filename, fields, trees,
                      max_file_size, update, save_roots_only)

    header_filename = save_header_file(
        arbor, filename, fields, root_field_data,
        group_nnodes, group_ntrees)

    return header_filename

def determine_save_state(arbor, filename, fields, trees, in_place):
    """
    Determine if we can just output new analysis fields to
    sidecar files and skip saving the rest.

    If updating return filenames associated with currently loaded arbor.
    """

    from ytree.frontends.ytree.arbor import YTreeArbor

    fstate = (False, filename)
    tstate = (True, arbor.parameter_filename)

    if not isinstance(arbor, YTreeArbor):
        return fstate

    if filename is not None:
        return fstate

    if fields not in [None, "all"]:
        return fstate

    if not in_place and trees is not None:
        if len(trees) != arbor.size:
            return fstate

        my_uids = arbor.arr([t["uid"] for t in arbor])
        if (arbor["uid"] != my_uids).any():
            return fstate

    return tstate

def determine_output_filename(path, suffix):
    """
    Figure out the output filename.
    """

    if path is None:
        path = 'arbor'

    if path.endswith(suffix):
        dirname = os.path.dirname(path)
        filename = path[:-len(suffix)]
    else:
        dirname = path
        filename = os.path.join(
            dirname, os.path.basename(path))
    ensure_dir(dirname)
    return filename

def determine_field_list(arbor, fields, update):
    """
    Get the list of fields to be saved.
    """

    if fields in [None, "all"]:
        # If this is an update, don't resave disk fields.
        field_list = set(arbor.analysis_field_list)

        # If not an update, save all the fields.
        if not update:
            field_list.update(arbor.field_list)

        # If a field has an alias, get that instead.
        fields = []
        for field in field_list:
            fields.extend(
                arbor.field_info[field].get("aliases", [field]))

    else:
        # If not saving all trees, add uid and desc_uid fields
        # to make sure the tree can be built.
        fields = list(set(fields + ["uid", "desc_uid"]))

    return fields

def get_output_fieldnames(fields):
    """
    Get filenames as they will be written to disk.
    """

    return [field.replace("/", "_") for field in fields]

def save_data_files(arbor, filename, fields, trees,
                    max_file_size, update, nodes_only):
    """
    Write all data files by grouping trees together.

    Return arrays of number of nodes and trees written to each file
    as well as a dictionary of root fields.

    If update is True, use the file layout of the arbor instead of
    calculating from max_file_size.
    """

    # Keep the trees we want to transplant separate.
    if trees is not None and update:
        save_trees = trees

    if trees is None or update:
        trees = arbor._yield_root_nodes(range(arbor.size))
        save_size = arbor.size
    else:
        save_size = len(trees)

    if update:
        # Transplant field data onto tree roots.
        save_roots = transplant_analysis_fields(
            arbor, save_trees, nodes_only)
        # If we have new analysis fields, they must be written for
        # the entire arbor.
        save_all = any([arbor.field_info[field]["type"] == "analysis"
                        for field in fields])
        file_sizes = np.diff(arbor._node_io._ei, prepend=0)
    else:
        save_roots = {}
        save_all = True

    # This is the set of files we need to save.
    if not save_all:
        save_files = set(np.digitize([tree._ai for tree in save_roots.values()], arbor._node_io._ei))

    root_field_data = {field: [] for field in fields}

    group_nnodes = []
    group_ntrees = []
    current_group = []
    cg_nnodes = 0
    cg_ntrees = 0

    def my_save(cg_number, cg_nnodes, cg_ntrees):
        group_nnodes.append(cg_nnodes)
        group_ntrees.append(cg_ntrees)

        if update:
            total_guess = file_sizes.size
        else:
            total_guess = int(np.round(save_size * (cg_number+1) / sum(group_ntrees)))

        # If we don't need to save the data file, just gather root fields.
        if not save_all and cg_number not in save_files:

            mylog.info(f"[{cg_number+1}] / [~{total_guess}]: Compiling root fields.")
            cg_start = current_group[0]._arbor_index
            cg_end = current_group[-1]._arbor_index + 1
            for field in fields:
                root_field_data[field].append(arbor[field][cg_start:cg_end])

        else:
            save_data_file(
                arbor, filename, fields,
                np.array(current_group), root_field_data,
                cg_number, total_guess)

    i = 0
    for tree in trees:
        if update:
            tree = save_roots.get(tree._arbor_index, tree)

        current_group.append(tree)
        cg_nnodes += tree.tree_size
        cg_ntrees += 1

        # if updating, use file sizes of loaded arbor
        if (update and len(current_group) == file_sizes[i]) or \
          cg_nnodes > max_file_size:
            my_save(i, cg_nnodes, cg_ntrees)
            current_group = []
            cg_nnodes = 0
            cg_ntrees = 0
            i += 1

    if current_group:
        my_save(i, cg_nnodes, cg_ntrees)

    group_nnodes = np.array(group_nnodes)
    group_ntrees = np.array(group_ntrees)

    return group_nnodes, group_ntrees, root_field_data

def save_data_file(arbor, filename, fields, tree_group,
                   root_field_data,
                   current_iteration, total_guess):
    """
    Write data file for a single group of trees.
    """

    fieldnames = get_output_fieldnames(fields)

    arbor._node_io_loop(
        arbor._node_io.get_fields,
        pbar=f"Getting fields [{current_iteration+1} / ~{total_guess}]",
        root_nodes=tree_group, fields=fields, root_only=False)

    main_fdata  = {}
    main_ftypes = {}

    analysis_fdata  = {}
    analysis_ftypes = {}

    my_tree_size  = np.array([tree.tree_size for tree in tree_group])
    my_tree_end   = my_tree_size.cumsum()
    my_tree_start = my_tree_end - my_tree_size
    for field, fieldname in zip(fields, fieldnames):
        fi = arbor.field_info[field]

        if fi.get("type") in ["analysis", "analysis_saved"]:
            my_fdata  = analysis_fdata
            my_ftypes = analysis_ftypes
        else:
            my_fdata  = main_fdata
            my_ftypes = main_ftypes

        my_ftypes[fieldname] = "data"
        my_fdata[fieldname]  = np.concatenate(
            [node.field_data[field] if node.is_root else node["tree", field]
             for node in tree_group])
        root_field_data[field].append(my_fdata[fieldname][my_tree_start])

    # In case we have saved any non-root trees,
    # mark them as having no descendents.
    if "desc_uid" in main_fdata:
        main_fdata["desc_uid"][my_tree_start] = -1

    for node in tree_group:
        arbor.reset_node(node)

    if main_fdata:
        main_fdata["tree_start_index"] = my_tree_start
        main_fdata["tree_end_index"]   = my_tree_end
        main_fdata["tree_size"]        = my_tree_size
        for ft in ["tree_start_index",
                   "tree_end_index",
                   "tree_size"]:
            main_ftypes[ft] = "index"
        my_filename = f"{filename}_{current_iteration:04d}.h5"
        save_as_dataset({}, my_filename, main_fdata,
                        field_types=main_ftypes)

    if analysis_fdata:
        my_filename = f"{filename}_{current_iteration:04d}-analysis.h5"
        save_as_dataset({}, my_filename, analysis_fdata,
                        field_types=analysis_ftypes)

def save_header_file(arbor, filename, fields, root_field_data,
                     group_nnodes, group_ntrees):
    """
    Write the header file.
    """

    ds = {}
    for attr in ["hubble_constant",
                 "omega_matter",
                 "omega_lambda"]:
        if getattr(arbor, attr, None) is not None:
            ds[attr] = getattr(arbor, attr)

    # Data structures for disk fields.
    main_fi     = {}
    main_rdata  = {}
    main_rtypes = {}

    # Analysis fields saved separately.
    analysis_fi     = {}
    analysis_rdata  = {}
    analysis_rtypes = {}

    fieldnames = get_output_fieldnames(fields)
    for field, fieldname in zip(fields, fieldnames):
        fi = arbor.field_info[field]

        if fi.get("type") in ["analysis", "analysis_saved"]:
            my_fi     = analysis_fi
            my_rdata  = analysis_rdata
            my_rtypes = analysis_rtypes
        else:
            my_fi     = main_fi
            my_rdata  = main_rdata
            my_rtypes = main_rtypes

        my_fi[fieldname] = \
          dict((key, fi[key])
               for key in ["units", "description"]
               if key in fi)
        my_rdata[fieldname] = np.concatenate(root_field_data[field])
        my_rtypes[fieldname] = "data"

    # all saved trees will be roots
    if "desc_uid" in main_rdata:
        main_rdata["desc_uid"][:] = -1

    # Save the primary fields.
    header_filename = f"{filename}.h5"
    if main_fi:
        tree_end_index   = group_ntrees.cumsum()
        tree_start_index = tree_end_index - group_ntrees

        extra_attrs = {
            "arbor_type": "YTreeArbor",
            "unit_registry_json": arbor.unit_registry.to_json(),
            "unit_system_name": arbor.unit_registry.unit_system.name}
        if arbor.box_size is not None:
            extra_attrs["box_size"] = arbor.box_size
        extra_attrs["field_info"] = json.dumps(main_fi)
        extra_attrs["total_files"] = group_nnodes.size
        extra_attrs["total_trees"] = group_ntrees.sum()
        extra_attrs["total_nodes"] = group_nnodes.sum()
        hdata = {"tree_start_index": tree_start_index,
                 "tree_end_index"  : tree_end_index,
                 "tree_size"       : group_ntrees}
        hdata.update(main_rdata)
        del main_rdata
        htypes = dict((f, "index") for f in hdata)
        htypes.update(main_rtypes)

        save_as_dataset(ds, header_filename, hdata,
                        field_types=htypes,
                        extra_attrs=extra_attrs)
        del hdata

    # Save analysis fields to a sidecar file.
    if analysis_fi:
        extra_attrs = {}
        extra_attrs["field_info"] = json.dumps(analysis_fi)
        hdata = analysis_rdata
        del analysis_rdata
        htypes = dict((f, "index") for f in hdata)
        htypes.update(analysis_rtypes)

        analysis_header_filename = f"{filename}-analysis.h5"
        save_as_dataset(ds, analysis_header_filename, hdata,
                        field_types=htypes,
                        extra_attrs=extra_attrs)
        del hdata

    return header_filename

def transplant_analysis_fields(arbor, old_trees, nodes_only):
    """
    Copy analysis field data from one set of trees to another.

    The new tree list is likely to be the list of all trees in the arbor.
    The purpose of this is to allow saving in place of a subset of trees
    (including non-rootes) back into the original arbor.
    """

    root_ids = {}

    pbar = get_pbar("Transplanting analysis fields", len(old_trees))
    for i, old_tree in enumerate(old_trees):
        old_root = old_tree.find_root()
        ai = old_root._arbor_index
        if ai in root_ids:
            new_root = root_ids[ai]
        else:
            new_root = arbor[ai]
            root_ids[ai] = new_root

        if nodes_only:
            indices = old_tree.tree_id
        elif old_tree.is_root:
            indices = slice(None)
        else:
            indices = (old_tree._tree_field_indices)

        arbor._node_io.get_fields(new_root, fields=arbor.analysis_field_list)
        for field in arbor.analysis_field_list:
            if nodes_only:
                my_field = field
            elif old_tree.is_root:
                my_field = ("forest", field)
            else:
                my_field = ("tree", field)
            new_root.field_data[field][indices] = old_tree[my_field]

        pbar.update(i+1)
    pbar.finish()

    return root_ids



"""
TreeFarmArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import \
    defaultdict
import glob
import h5py
import re

from unyt.unit_registry import \
    UnitRegistry

from ytree.data_structures.arbor import \
    CatalogArbor
from ytree.frontends.treefarm.fields import \
    TreeFarmFieldInfo
from ytree.frontends.treefarm.io import \
    TreeFarmDataFile
from ytree.utilities.io import \
    _hdf5_yt_attr, \
    parse_h5_attr

class TreeFarmArbor(CatalogArbor):
    """
    Class for Arbors created with :class:`~ytree.treefarm.TreeFarm`.
    """

    _suffix = ".h5"
    _field_info_class = TreeFarmFieldInfo
    _data_file_class = TreeFarmDataFile

    def _parse_parameter_file(self):
        fh = h5py.File(self.filename, "r")

        for attr in ["hubble_constant",
                     "omega_matter",
                     "omega_lambda"]:
            setattr(self, attr, fh.attrs[attr])

        my_ur = UnitRegistry.from_json(
            parse_h5_attr(fh, "unit_registry_json"))
        right = _hdf5_yt_attr(fh, "domain_right_edge",
                              unit_registry=my_ur)
        left  = _hdf5_yt_attr(fh, "domain_left_edge",
                              unit_registry=my_ur)
        # Drop the "cm" suffix because all lengths will
        # be in comoving units.
        self.box_size = self.quan(
            (right - left)[0].to("Mpccm/h"), "Mpc/h")
        fh.close()

    def _get_data_files(self):
        suffix = ".0" + self._suffix
        reg = re.search(rf"^(.+\D)\d+{suffix}$", self.filename)
        if reg is None:
            raise RuntimeError(
                f"Cannot determine numbering system for {self.filename}.")
        prefix = reg.groups()[0]
        files = glob.glob(f"{prefix}*{self._suffix}")
        fids = defaultdict(list)
        for my_file in files:
            fid = int(my_file[len(prefix):-len(suffix)])
            fids[fid].append(my_file)
        my_files = [fids[myfid] for myfid in sorted(fids.keys(), reverse=True)]
        self.data_files = [[self._data_file_class(f, self)
                            for f in fl] for fl in my_files]

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .h5, be loadable as an hdf5 file,
        and have a "data_type" attribute.
        """
        fn = args[0]
        if not fn.endswith(self._suffix):
            return False
        try:
            with h5py.File(fn, "r") as f:
                dtype = f.attrs.get("data_type")
                if isinstance(dtype, bytes):
                    dtype = dtype.astype(str)
                if dtype != "halo_catalog":
                    return False
        except BaseException:
            return False
        return True



"""
tests for example scripts



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.utilities.testing import \
    ExampleScriptTest, \
    TempDirTest

class TestPlotMostMassive(TempDirTest, ExampleScriptTest):
    script_filename = "plot_most_massive.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = ("most_massive.png",)

class TestPlotMostHalos(TempDirTest, ExampleScriptTest):
    script_filename = "plot_most_halos.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = ("most_halos.png",)

class TestHaloAge(TempDirTest, ExampleScriptTest):
    ncores = 2
    script_filename = "halo_age.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = (
        "halo_age/halo_age-analysis.h5",
        "halo_age/halo_age.h5",
        "halo_age/halo_age_0000-analysis.h5",
        "halo_age/halo_age_0000.h5"
    )

class TestHaloSignificance(TempDirTest, ExampleScriptTest):
    ncores = 2
    script_filename = "halo_significance.py"
    input_filename = "tiny_ctrees/locations.dat"
    output_files = (
        "halo_significance/halo_significance-analysis.h5",
        "halo_significance/halo_significance.h5",
        "halo_significance/halo_significance_0000-analysis.h5",
        "halo_significance/halo_significance_0000.h5"
    )



"""
Alter the xml file produced by pytest so the file attribute
reflects the script file and not the file of the base class.
"""

import os
import sys
import xml.etree.ElementTree as ET

if __name__ == "__main__":
    ifn = sys.argv[1]
    print (f"Reading {ifn}.")
    tree = ET.parse(ifn)
    root = tree.getroot()

    for child in root[0]:
        fn = child.attrib["file"]
        if fn.startswith("tests"):
            continue

        cln = child.attrib["classname"]
        fnew = os.path.join(*cln.split(".")[:-1]) + ".py"
        print (f"Changing {fn} to {fnew}.")
        child.attrib["file"] = fnew

    print (f"Writing {ifn}.")
    tree.write(ifn)



"""
TreeFarmArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from ytree.data_structures.fields import \
    FieldInfoContainer

m_unit = "Msun"
p_unit = "unitary"
r_unit = "kpc"
v_unit = "km/s"

class TreeFarmFieldInfo(FieldInfoContainer):
    alias_fields = (
        ("halo_id", "particle_identifier", None),
        ("desc_id", "descendent_identifier", None),
        ("mass", "particle_mass", m_unit),
        ("position_x", "particle_position_x", p_unit),
        ("position_y", "particle_position_y", p_unit),
        ("position_z", "particle_position_z", p_unit),
        ("velocity_x", "particle_velocity_x", v_unit),
        ("velocity_y", "particle_velocity_y", v_unit),
        ("velocity_z", "particle_velocity_z", v_unit),
    )



"""
Arbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import \
    defaultdict
import functools
import numpy as np
import os

from unyt import \
    unyt_array, \
    unyt_quantity

from yt.funcs import \
    get_pbar, \
    TqdmProgressBar
from unyt.dimensions import \
    dimensionless, \
    length
from unyt.unit_registry import \
    UnitRegistry
from yt.utilities.cosmology import \
    Cosmology

from ytree.data_structures.detection import \
    SelectionDetector
from ytree.data_structures.fields import \
    FieldContainer, \
    FieldInfoContainer
from ytree.data_structures.io import \
    DefaultRootFieldIO, \
    TreeFieldIO
from ytree.data_structures.node_link import \
    NodeLink
from ytree.data_structures.save_arbor import \
    save_arbor
from ytree.data_structures.node_container import NodeContainer
from ytree.data_structures.tree_node import \
    TreeNode
from ytree.data_structures.tree_node_selector import \
    tree_node_selector_registry
from ytree.utilities.logger import \
    ytreeLogger, \
    fake_pbar

arbor_registry = {}
_selection_types = ("forest", "tree", "prog")

class RegisteredArbor(type):
    """
    Add to the registry of known Arbor classes to cycle
    through in the load function.
    """
    def __init__(cls, name, b, d):
        type.__init__(cls, name, b, d)
        arbor_type = name[:name.rfind("Arbor")]
        if arbor_type:
            arbor_registry[arbor_type] = cls

class Arbor(metaclass=RegisteredArbor):
    """
    Base class for all Arbor classes.

    Loads a merger tree output file or a series of halo catalogs
    and create trees, stored in an array in
    :func:`~ytree.data_structures.arbor.Arbor.trees`.
    Arbors can be saved in a universal format with
    :func:`~ytree.data_structures.arbor.Arbor.save_arbor`. Also, provide some
    convenience functions for creating unyt_arrays and unyt_quantities and
    a cosmology calculator.
    """

    _field_info_class = FieldInfoContainer
    _root_field_io_class = DefaultRootFieldIO
    _tree_field_io_class = TreeFieldIO
    _default_dtype = np.float64

    ### attributes required for generating a TreeNode object
    ### for a given Arbor class.
    ### We store these in arrays and use them to generate TreeNodes
    ### when they are needed.
    ## attributes required for constructing TreeNodes
    _node_con_attrs = ('uid',)
    ## attributes we may not have, but would be nice if we did
    _node_too_attrs = ('_tree_size',)
    ## attributes specific to an Arbor class for facilitating io
    _node_io_attrs = ()

    ### tree node attributes for all Arbor types.
    ### These facilitate walking the tree, getting fields, etc.
    ### We keep track of these for resetting TreeNodes and
    ### deciding when they are setup or grown.
    _reset_attrs = ("_tfi", "_pfi")
    _setup_attrs = ("_desc_uids", "_uids")
    _grow_attrs = ("_link_storage", "_link")

    omega_matter = None
    omega_lambda = None
    omega_radiation = 0

    def __init__(self, filename):
        """
        Initialize an Arbor given an input file.
        """

        self._set_paths(filename)
        self._parse_parameter_file()
        self._set_units()
        self._setup_io()
        self._get_data_files()
        self._setup_fields()
        self._set_default_selector()

    def _set_paths(self, filename):
        """
        Set data paths.
        """

        self.filename = filename
        if isinstance(filename, (list, tuple)):
            fn = filename[0]
        else:
            fn = filename
        self._set_parameter_filename(fn)
        self.basename = os.path.basename(fn)
        dn = os.path.dirname(fn)
        self.directory = dn if dn else '.'

    def _set_parameter_filename(self, filename):
        """
        Set the name of the file from which to get metadata.
        """

        self.parameter_filename = filename

    def _parse_parameter_file(self):
        """
        Read relevant parameters from parameter file or file header
        and detect fields.
        """
        raise NotImplementedError

    def _set_units(self):
        """
        Set "cm" units for explicitly comoving.
        Note, we are using comoving units all the time since
        we are dealing with data at multiple redshifts.
        """
        for my_unit in ["m", "pc", "AU"]:
            new_unit = f"{my_unit}cm"
            self.unit_registry.add(
                new_unit, self.unit_registry.lut[my_unit][0],
                length, self.unit_registry.lut[my_unit][3])

    @functools.cached_property
    def cosmology(self):
        """
        Cosmology calculator with self-consistent parameters and units.

        See :ref:`yt:cosmology-calculator` for more information.
        """

        missing = []
        for attr in ["hubble_constant",
                     "omega_matter",
                     "omega_lambda"]:
            if getattr(self, attr) is None:
                missing.append(attr)

        if missing:
            ytreeLogger.warning(
                "Cannot create cosmology calculator without missing attributes: " +
                str(missing))
            return None

        else:
            return Cosmology(
                hubble_constant=self.hubble_constant,
                omega_matter=self.omega_matter,
                omega_lambda=self.omega_lambda,
                omega_radiation=self.omega_radiation,
                unit_registry=self.unit_registry)

    def _setup_io(self):
        """
        Create field io objects.
        """
        self._node_io = self._tree_field_io_class(
            self, default_dtype=self._default_dtype)
        self._root_io = self._root_field_io_class(
            self, default_dtype=self._default_dtype)

    def _get_data_files(self):
        """
        Get all files that hold field data and make them known
        to the i/o system.
        """
        pass

    def _setup_fields(self):
        """
        Setup field containers and definitions.
        """
        self.field_data = FieldContainer(self)
        self.derived_field_list = []
        # This may already be created by ytree frontend.
        if not hasattr(self, "analysis_field_list"):
            self.analysis_field_list = []
        self.field_info.setup_known_fields()
        self.field_info.setup_aliases()
        self.field_info.setup_derived_fields()
        self.field_info.setup_vector_fields()

    def _set_default_selector(self):
        """
        Set the default tree node selector as maximum mass.
        """
        self.set_selector("max_field_value", "mass")

    @property
    def is_planted(self):
        """
        Determine if trees have been planted.
        """
        return self._node_info_storage is not None

    def _plant_trees(self):
        """
        Create arrays to construct root nodes.
        """
        raise NotImplementedError

    _node_info_storage = None
    @property
    def _node_info(self):
        """
        The dict of arrays for storing node information.
        """

        if self._node_info_storage is not None:
            return self._node_info_storage
        self._initialize_node_info()
        return self._node_info_storage

    def _initialize_node_info(self):
        """
        Initialize the node_info arrays.
        """

        attrs = self._node_con_attrs + \
          self._node_io_attrs

        self._node_info_storage = \
          dict((attr, np.empty(self._size, dtype=np.int64))
               for attr in attrs)
        # initialize the target of opportunity attributes
        self._node_info_storage.update(
            dict((attr, -np.ones(self._size, dtype=np.int64))
                for attr in self._node_too_attrs))

    def is_setup(self, tree_node):
        """
        Return True if arrays of uids and descendent uids have
        been read in. Setup has also completed if tree is already
        grown.
        """
        return self.is_grown(tree_node) or \
          tree_node._uids is not None

    def _setup_tree(self, tree_node, **kwargs):
        """
        Create arrays of uids and desc_uids and attach them to the
        root node.
        """

        # skip if this is not a root or if already setup
        if self.is_setup(tree_node):
            return

        idtype = np.int64
        fields, _ = \
          self.field_info.resolve_field_dependencies(["uid", "desc_uid"])
        halo_id_f, desc_id_f = fields
        dtypes = {halo_id_f: idtype, desc_id_f: idtype}
        # Note to self, we call _read_fields and not _get_fields to
        # avoid recursion issues.
        field_data  = self._node_io._read_fields(tree_node, fields,
                                                 dtypes=dtypes, **kwargs)

        tree_node._uids      = field_data[halo_id_f]
        tree_node._desc_uids = field_data[desc_id_f]
        tree_node._tree_size = tree_node._uids.size
        tree_node.field_data["uid"] = tree_node._uids
        tree_node.field_data["desc_uid"] = tree_node._desc_uids

    def is_grown(self, tree_node):
        """
        Return True if a tree has been fully assembled, i.e.,
        the hierarchy of ancestor tree nodes has been built.
        """
        return tree_node.root != -1

    def _grow_tree(self, tree_node, **kwargs):
        """
        Construct the hierarchy of ancestors and descendents
        for all nodes in the tree.
        """

        # skip this if not a root or if already grown
        if self.is_grown(tree_node):
            return

        self._setup_tree(tree_node, **kwargs)
        size      = tree_node.tree_size
        uids      = tree_node.uids
        desc_uids = tree_node.desc_uids
        links     = np.empty(size, dtype=object)

        # Make a dict mapping uids to index of storage array.
        # First, try to get indices out as the dict is constructed
        # since the dict will be smaller at first.
        uidmap = {}
        not_found = []
        for i, (uid, desc_uid) in enumerate(zip(uids, desc_uids)):
            node = NodeLink(i)
            uidmap[uid] = i
            desc_index = uidmap.get(desc_uid)
            if desc_index is None:
                not_found.append((node, desc_uid))
            else:
                desc = links[desc_index]
                desc.add_ancestor(node)
            links[i] = node

        # Make any additional links missed on the first pass.
        for node, desc_uid in not_found:
            if desc_uid == -1:
                continue
            desc = links[uidmap[desc_uid]]
            desc.add_ancestor(node)

        tree_node.root = tree_node
        tree_node._link = links[0]
        tree_node._link_storage = links

    _attr_map = None
    def _build_attr(self, attr, tree_node):
        """
        Call the correct function for building a given attribute.
        """

        if self._attr_map is None:
            self._attr_map = \
              dict([(attr, self._setup_tree)
                    for attr in self._setup_attrs] +
                   [(attr, self._grow_tree)
                    for attr in self._grow_attrs])

        self._attr_map[attr](tree_node)

    def reset_node(self, tree_node):
        """
        Reset all data structures for a single node.

        The goal is to clear as many data structures as
        possible without rendering the node object useless,
        if they are intended to be kept around.

        Calling reset_node on a non-root node should not make
        the non-root node useless.

        Calling reset_node on a root node will render any
        generated non-root nodes useless.
        """

        tree_node.clear_fields()
        attrs = self._reset_attrs

        if tree_node.is_root:
            if self.is_grown(tree_node):
                attrs += self._grow_attrs
                tree_node.root = -1
            if self.is_setup(tree_node):
                attrs += self._setup_attrs

        for attr in attrs:
            setattr(tree_node, attr, None)

    @property
    def ytds(self):
        raise NotImplementedError(
            "Only ytree data can be loaded with yt. "
            "Save data with save_arbor and then reload.")

    def _node_io_loop(self, func, *args, **kwargs):
        """
        Call the provided function over a list of nodes.

        If possible, group nodes by common data files to speed
        things up.

        Parameters
        ----------
        func : function
            Function to be called on an array of nodes.
        pbar : optional, string or yt.funcs.TqdmProgressBar
            A progress bar to be updated with each iteration.
            If a string, a progress bar will be created and the
            finish function will be called. If a progress bar is
            provided, the finish function will not be called.
            Default: None (no progress bar).
        root_nodes : optional, array of root TreeNodes
            Array of nodes over which the function will be called.
            If None, the list will be self[:] (i.e., all
            root_nodes).
            Default: None.

        Returns
        -------
        rvals : list
            return values from calling func on each node.
            These will have the same order as the original node list.
        """

        self._plant_trees()

        pbar = kwargs.pop("pbar", None)
        root_nodes = kwargs.pop("root_nodes", None)

        data_files, node_list, return_order = \
          self._node_io_loop_prepare(root_nodes)
        nnodes = sum([nodes.size for nodes in node_list])

        finish = True
        if pbar is None:
            pbar = fake_pbar("", nnodes)
        elif not isinstance(pbar, TqdmProgressBar):
            pbar = get_pbar(pbar, nnodes)
        else:
            finish = False

        rvals = []
        c = 0
        for data_file, nodes in zip(data_files, node_list):
            self._node_io_loop_start(data_file)

            # if we're doing all of them, just give the indices
            if root_nodes is None:
                my_nodes = nodes
            else:
                my_nodes = root_nodes[nodes]

            for node in self._yield_root_nodes(my_nodes):
                rval = func(node, *args, **kwargs)
                rvals.append(rval)
                c += 1
                pbar.update(c)

            self._node_io_loop_finish(data_file)

        if finish:
            pbar.finish()

        if return_order is not None:
            rvals = [rvals[i] for i in return_order]

        return rvals

    def _node_io_loop_start(self, data_file):
        pass

    def _node_io_loop_finish(self, data_file):
        pass

    def _node_io_loop_prepare(self, nodes):
        """
        This is called at the beginning of _node_io_loop.

        In different frontends, this can be used to group nodes by
        common data files. If nodes is None, we want all root
        nodes in the Arbor.

        Below is the default behavior, which does the bare minimum
        of returning:
        list of [None] : meaning all nodes come in a single group
            associated with no particular data file.
        list containing array of all provided nodes: meaning there
            is no specific grouping to be done
        None : meaning the nodes do not have to be reordered after
            being processed.

        See the implementation in individual frontends for
        more informative examples.

        Returns
        -------
        data_files : list
            list of data files that will be used
        index_list : list of arrays
            indices of the provided array of nodes associated
            with each of the data files
        return_order : int array
            array of indices used to reorder the return values
            to the order of the provided nodes
        """

        self._plant_trees()

        if nodes is None:
            my_size = self.size
        else:
            my_size = nodes.size
        indices = np.arange(my_size)

        return [None], [indices], None

    def __iter__(self):
        """
        Iterate over all trees in the arbor.
        """

        self._plant_trees()
        for node in self._yield_root_nodes(range(self.size)):
            yield node

    def __repr__(self):
        return self.basename

    def __getitem__(self, key):
        return self.query(key)

    def query(self, key):
        """
        If given a string, return an array of field values for the
        roots of all trees.
        If given an integer, return a tree from the list of trees.
        """

        if isinstance(key, str):
            if key in _selection_types:
                raise SyntaxError("Argument must be a field or integer.")
            self._root_io.get_fields(self, fields=[key])
            return self.field_data[key]
        return self._generate_root_nodes(key)

    def _generate_root_nodes(self, key):
        """
        Create root nodes given an index or slice from uid array.
        """

        self._plant_trees()
        if isinstance(key, (int, np.integer)):
            return self._generate_root_node(key)
        elif isinstance(key, slice) or isinstance(key, np.ndarray):
            indices = np.arange(self.size)[key]
            return self._yield_root_nodes(indices)
        else:
            raise ValueError('Cannot generate nodes from argument: ', key)

    def _yield_root_nodes(self, indices):
        """
        Root node generator.
        """

        # If we've been given an array of TreeNodes,
        # just yield them back.
        if getattr(indices, 'dtype', None) == object:
            for index in indices:
                yield index
            return

        for index in indices:
            node = self._generate_root_node(index)
            yield node

    def _generate_root_node(self, index):
        """
        Create a root node given its index in the array of uids.
        """

        args = tuple(self._node_info[attr][index]
                      for attr in self._node_con_attrs)
        my_node = TreeNode(*args, arbor=self, root=True)
        my_node._arbor_index = index

        for attr in self._node_io_attrs:
            setattr(my_node, attr, self._node_info[attr][index])

        for attr in self._node_too_attrs:
            val = self._node_info[attr][index]
            if val != -1:
                setattr(my_node, attr, self._node_info[attr][index])

        return my_node

    def _generate_tree_node(self, root_node, node_link):
        """
        Create a non-root node in a tree.
        """

        tree_id = node_link.tree_id
        if tree_id == 0:
            return root_node
        uid        = root_node.uids[tree_id]
        node       = TreeNode(uid, arbor=self, root=False)
        node.root  = root_node
        node._link = node_link
        return node

    def _store_node_info(self, tree_node, attr):
        """
        Store a TreeNode attribute an array for retrieval later.
        """

        self._node_info[attr][tree_node._arbor_index] = \
          getattr(tree_node, attr)

    _field_info = None
    @property
    def field_info(self):
        """
        A dictionary containing information for each available field.
        """
        if self._field_info is None and \
          self._field_info_class is not None:
            self._field_info = self._field_info_class(self)
        return self._field_info

    _size = None
    @property
    def size(self):
        """
        Return total number of trees.
        """
        if self._size is None:
            self._plant_trees()
        return self._size

    def __len__(self):
        """
        Return total number of trees.
        """
        return self.size

    _unit_registry = None
    @property
    def unit_registry(self):
        """
        Unit system registry.
        """
        if self._unit_registry is None:
            self._unit_registry = UnitRegistry()
        return self._unit_registry

    @unit_registry.setter
    def unit_registry(self, value):
        self._unit_registry = value
        self._arr = None
        self._quan = None

    _hubble_constant = None
    @property
    def hubble_constant(self):
        """
        Value of the Hubble parameter.
        """
        return self._hubble_constant

    @hubble_constant.setter
    def hubble_constant(self, value):
        self._hubble_constant = value
        if value is None:
            return
        # reset the unit registry lut while preserving other changes
        self.unit_registry = UnitRegistry.from_json(
            self.unit_registry.to_json())
        if 'h' in self.unit_registry:
            self.unit_registry.modify("h", self.hubble_constant)
        else:
            self.unit_registry.add(
                'h', self.hubble_constant, dimensionless)

    _box_size = None
    @property
    def box_size(self):
        """
        The simulation box size.
        """
        return self._box_size

    @box_size.setter
    def box_size(self, value):
        self._box_size = value
        # set unitary as soon as we know the box size
        self.unit_registry.add(
            "unitary", float(self.box_size.in_base()), length)

    def set_selector(self, selector, *args, **kwargs):
        r"""
        Sets the tree node selector to be used.

        This sets the manner in which halo progenitors are
        chosen from a list of ancestors.  The most obvious example
        is to select the most massive ancestor.

        Parameters
        ----------
        selector : string
            Name of the selector to be used.

        Any additional arguments and keywords to be provided to
        the selector function should follow.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("rockstar_halos/trees/tree_0_0_0.dat")
        >>> a.set_selector("max_field_value", "mass")

        """
        self.selector = tree_node_selector_registry.find(
            selector, *args, **kwargs)

    _arr = None
    @property
    def arr(self):
        """
        Create a unyt_array using the Arbor's unit system.

        Parameters
        ----------

        input_array : Iterable
            A tuple, list, or array to attach units to
        units: String unit specification, unit symbol or astropy object
            The units of the array. Powers must be specified using python syntax
            (cm**3, not cm^3).
        dtype : string or NumPy dtype object
            The dtype of the returned array data

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("consistent_trees/tree_0_0_0.dat")
        >>> a.arr([1,2,3], "kg")
        unyt_array([1, 2, 3], 'kg')

        """
        if self._arr is not None:
            return self._arr
        self._arr = functools.partial(unyt_array,
                                      registry=self.unit_registry)
        return self._arr

    _quan = None
    @property
    def quan(self):
        """
        Create a unyt_quantity using the Arbor's unit system.

        Parameters
        ----------

        input_array : Iterable
            A tuple, list, or array to attach units to
        units: String unit specification, unit symbol or astropy object
            The units of the array. Powers must be specified using python syntax
            (cm**3, not cm^3).
        dtype : string or NumPy dtype object
            The dtype of the returned array data

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("consistent_trees/tree_0_0_0.dat")
        >>> a.quan(34, "cm")
        unyt_quantity(34, 'cm')

        """
        if self._quan is not None:
            return self._quan
        self._quan = functools.partial(unyt_quantity,
                                       registry=self.unit_registry)
        return self._quan

    @functools.cached_property
    def container(self):
        """
        Helper function for creating
        :class:`~ytree.data_structures.node_container.NodeContainer` objects.

        See :class:`~ytree.data_structures.node_container.NodeContainer` for more
        information.
        """
        nc = functools.partial(NodeContainer, arbor=self)
        nc.__doc__ = NodeContainer.__doc__
        return nc

    def select_halos(self, criteria, trees=None,
                     select_from=None, fields=None):
        """
        Select halos from the arbor based on a set of criteria given as a string.

        Halos matching the criteria will be returned through a generator. Matches
        are returned as soon as they are found, allowing you to begin working
        with them before the search has completed. The progress bar will update
        to report the number of matches found as the search progresses.

        Parameters
        ----------

        criteria : string
            A string that will eval to a Numpy-like selection operation
            performed on a TreeNode object called "tree".
            Example: 'tree["tree", "redshift"] > 1'
        trees : optional, list or array of TreeNodes
            A list or array of TreeNode objects in which to search. If none given,
            the search is performed over the full arbor.
        select_from : deprecated, do not use
            This keyword is no longer required and using it does nothing.
        fields : deprecated, do not use
            This keyword is no longer required and using it does nothing.

        Returns
        -------

        halos : :class:`~ytree.data_structures.tree_node.TreeNode` generator
            A generator yielding all TreeNodes meeting the criteria.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> for halo in a.select_halos('tree["tree", "redshift"] > 1'):
        ...     print (halo["mass"])
        >>>
        >>> halos = list(a.select_halos('tree["prog", "mass"].to("Msun") >= 1e10'))
        >>> print (len(halos))

        """

        if select_from is not None:
            import warnings
            from numpy.exceptions import VisibleDeprecationWarning
            warnings.warn(
                "The \"select_from\" keyword is deprecated and no longer does anything.",
                VisibleDeprecationWarning, stacklevel=2)

        if fields is not None:
            import warnings
            from numpy.exceptions import VisibleDeprecationWarning
            warnings.warn(
                "The \"fields\" keyword is deprecated and no longer does anything.",
                VisibleDeprecationWarning, stacklevel=2)

        tree = SelectionDetector(self)
        eval(criteria)
        if len(tree.selectors) > 1:
            raise ValueError(
                f"Selection criteria must only use one selector: \"{criteria}\".\n"
                f"    Selection criteria uses {len(tree.selectors)} selectors: "
                f"{tree.selectors}.")
        selector = tree.selectors[0]

        if trees is None:
            trees = self

        found = 0
        pbar = get_pbar(f"Selecting halos ({found} found)", trees.size)
        for i, tree in enumerate(trees):
            imatches = np.where(eval(criteria))[0]
            if imatches.size > 0:
                found += imatches.size
                if isinstance(pbar, TqdmProgressBar):
                    pbar._pbar.set_description_str(f"Selecting halos (found {found})")
            pbar.update(i+1)

            for imatch in imatches:
                yield tree.get_node(selector, imatch)

        pbar.finish()

    def add_analysis_field(self, name, units, dtype=None, default=0):
        r"""
        Add an empty field to be filled by analysis operations.

        Parameters
        ----------
        name : string
            Field name.
        units : string
            Field units.
        dtype : optional, type
            Data type for field values. If None, the default data type
            of the arbor is used.
            Default: None.
        default: optional, numeric
            Default field value when field is initialized.
            Default: 0.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> a.add_analysis_field("robots", "Msun * kpc")
        >>> # Set field for some halo.
        >>> my_tree = a[0]
        >>> my_tree["tree"][7]["robots"] = 1979.816
        """

        self.field_info.add_analysis_field(
            name, units,
            dtype=dtype, default=default)

    def add_alias_field(self, alias, field, units=None,
                        force_add=True):
        r"""
        Add a field as an alias to another field.

        Parameters
        ----------
        alias : string
            Alias name.
        field : string
            The field to be aliased.
        units : optional, string
            Units in which the field will be returned.
        force_add : optional, bool
            If True, add field even if it already exists and warn the
            user and raise an exception if dependencies do not exist.
            If False, silently do nothing in both instances.
            Default: True.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> # "Mvir" exists on disk
        >>> a.add_alias_field("mass", "Mvir", units="Msun")
        >>> print (a["mass"])

        """

        self.field_info.add_alias_field(
            alias, field, units=units, force_add=force_add)

    def add_derived_field(self, name, function,
                          units=None, dtype=None, description=None,
                          vector_components=None, force_add=True):
        r"""
        Add a field that is a function of other fields.

        Parameters
        ----------
        name : string
            Field name.
        function : callable
            The function to be called to generate the field.
            This function should take two arguments, the
            arbor and the data structure containing the
            dependent fields.  See below for an example.
        units : optional, string
            The units in which the field will be returned.
        dtype : optional, type
            The data type of the field array. If none, use the
            default type set by Arbor._default_dtype.
        description : optional, string
            A short description of the field.
        vector_components : optional, list of strings
            If given, the field is assumed to be a vector field with
            the x/y/z component fields provided here.
            Default: None.
        force_add : optional, bool
            If True, add field even if it already exists and warn the
            user and raise an exception if dependencies do not exist.
            If False, silently do nothing in both instances.
            Default: True.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> def _redshift(field, data):
        ...     return 1. / data["scale"] - 1
        ...
        >>> a.add_derived_field("redshift", _redshift)
        >>> print (a["redshift"])

        """

        self.field_info.add_derived_field(
            name, function,
            units=units, dtype=dtype, description=description,
            vector_components=vector_components, force_add=force_add)

    def add_vector_field(self, name, vector_components=None):
        """
        Add vector fields for a set of x,y,z component fields.

        This will add a general vector field that returns the combined
        x, y, z components as a single Nx3 array. A <field>_magnitude
        field with the quadrature sum of the components is also added.

        Parameters
        ----------
        name : string
            The name of the field. Component x,y,z fields must exist.
        vector_components : list of strings
            Used to explicitly provide the x/y/z component fields. If not
            given, a naming convention of <name>_<xyz> is assumed.
            Default: None.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("tree_0_0_0.dat")
        >>> for ax in 'xyz':
        >>>     a.add_analysis_field(f"thing_{ax}")
        >>> fn = a.save_arbor()
        >>> a_new = ytree.load(fn)
        >>> a_new.add_vector_field("thing")
        >>> print (a_new["thing"])
        >>> print (a_new["thing_magnitude"])

        >>> import ytree
        >>> a = ytree.load("AHF_100_tiny/GIZMO-NewMDCLUSTER_0047.snap_128.parameter")
        >>> a.add_vector_field("mean_z", vector_components=["mean_z_gas", "mean_z_star"])

        """

        self.field_info.add_vector_field(name, cfields=vector_components)

    def get_yt_selection(self, *args, **kwargs):
        raise NotImplementedError(
            "This function is only implemented for ytree arbors."
            "Use save_arbor to save your data in the correct format.")

    def get_nodes_from_selection(self, *args, **kwargs):
        raise NotImplementedError(
            "This function is only implemented for ytree arbors."
            "Use save_arbor to save your data in the correct format.")

    def reload_arbor(self):
        raise NotImplementedError(
            "This function is only implemented for ytree arbors."
            "Use save_arbor to save your data in the correct format.")

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        """
        Check if input file works with a specific Arbor class.
        This is used with :func:`~ytree.data_structures.arbor.load` function.
        """
        return False

    def save_arbor(self, **kwargs):
        r"""
        Save the arbor to a file.

        The saved arbor can be re-loaded as an arbor.

        Parameters
        ----------
        filename : optional, string
            Output file keyword.  If filename ends in ".h5",
            the main header file will be just that.  If not,
            filename will be <filename>/<basename>.h5.
            Default: "arbor".
        fields : optional, list of strings
            The fields to be saved.  If not given, all
            fields will be saved.
        trees : optional, list or array of TreeNodes
            If given, only save trees stemming from these nodes.
            If not provide, all trees will be saved.
        save_in_place : optional, bool or None
            If True, analysis fields will be saved to the original
            arbor, even if only a subset of all trees is provided
            with the trees keyword. This will essentially "update"
            the arbor in place. If False and only a subset of
            all trees is provided, a new arbor will be created
            containing only the trees provided. If set to None,
            behavior is determined by the type of arbor loaded.
            If the arbor is a YTreeArbor (i.e., saved with
            save_arbor), save_in_place will be set to True. If
            not of this type, it will be set to False.
            Default: None
        save_roots_only : optional, bool
            If True, only field values of each node are saved.
            If False, field data for the entire tree stemming
            from that node are saved.
            Default:  False.
        max_file_size : optional, float
            The maximum number of nodes saved to a single file.
            Smaller numbers will result in more files. Performance
            may change somewhat with different values.
            Default: 524288 (2^19).

        Returns
        -------
        header_filename : string
            The filename of the saved arbor.

        Examples
        --------

        >>> import ytree
        >>> a = ytree.load("rockstar_halos/trees/tree_0_0_0.dat")
        >>> fn = a.save_arbor()
        >>> # reload it
        >>> a2 = ytree.load(fn)

        """

        fn = save_arbor(self, **kwargs)
        return fn

class SegmentedArbor(Arbor):
    """
    Arbor subclass for multi-file datasets where an entire merger tree
    is contained within a file (i.e., no overlap). This permits the
    definition of a useful _node_io_loop_prepare function.
    """

    # Data formats organized similar to below can use this class.
    # _fi - file index, i.e., which data file is it in
    # _si - start index, the array index where this tree starts
    _node_io_attrs = ('_fi', '_si')

    def _node_io_loop_start(self, data_file):
        data_file.open()

    def _node_io_loop_finish(self, data_file):
        data_file.close()

    def _node_io_loop_prepare(self, nodes):
        if nodes is None:
            nodes = np.arange(self.size)
            fi = self._node_info['_fi']
            si = self._node_info['_si']
        elif nodes.dtype == object:
            fi = np.array(
                [node._fi if node.is_root else node.root._fi
                 for node in nodes])
            si = np.array(
                [node._si if node.is_root else node.root._si
                 for node in nodes])
        else: # assume an array of indices
            fi = self._node_info['_fi'][nodes]
            si = self._node_info['_si'][nodes]

        # the order they will be processed
        io_order = np.lexsort((si, fi))
        fi = fi[io_order]
        # array to return them to original order
        return_order = np.empty_like(io_order)
        return_order[io_order] = np.arange(io_order.size)

        ufi = np.unique(fi)
        data_files = [self.data_files[i] for i in ufi]
        index_list = [io_order[fi == i] for i in ufi]

        return data_files, index_list, return_order

class CatalogArbor(Arbor):
    """
    Base class for Arbors created from a series of halo catalog
    files where the descendent ID for each halo has been
    pre-determined.

    Unlike formats where tree information is stored in single file,
    halos are scattered about multiple catalog files. This requires
    us to store the root TreeNode objects and their full assemblies.
    """

    _prefix = None
    _data_file_class = None
    # does the dataset define unique ids?
    _has_uids = False

    # We will store root TreeNodes instead of generate them,
    # so we don't need to store anything here.
    _node_con_attrs = ()

    # Don't reset _ancestors or descendents because we won't be able to
    # rebuild trees without calling _plant_trees again.
    _setup_attrs = ("_desc_uids", "_uids", "_nodes", "_link_storage")
    _grow_attrs = ()

    def __init__(self, filename):
        super().__init__(filename)
        if not self._has_uids:
            if "uid" not in self.field_list:
                for field in "uid", "desc_uid":
                    self.field_list.append(field)
                    self.field_info[field] = {"units": "",
                                              "source": "arbor"}

    def _get_data_files(self):
        raise NotImplementedError

    def _generate_root_node(self, index):
        """
        Return a node from self._trees.

        These cannot be generated easily, so we keep them.
        """
        node = self._trees[index]
        if not hasattr(node, '_arbor_index'):
            node._arbor_index = index
        return node

    _trees = None
    @property
    def is_planted(self):
        """
        Determine if trees have been planted.
        """
        return self._trees is not None

    def _plant_trees(self):
        """
        Construct all trees.

        Since nodes are spread out over multiple files, we will
        plant all trees and create all ancestor/descendent links.

        The links will be held by the nodes themselves and we will
        not store the nodes in an array until _setup_tree is called.
        """

        if self.is_planted:
            return

        # This is a somewhat hacky way of catching halos with links
        # spanning more than one data set. That said, dict access
        # is much faster than I thought and perhaps this whole
        # routine needs to be refactored.
        if self._has_uids:
            all_dict = {}
            missed_connections = []

        # this can be called once with the list, but fields are
        # not guaranteed to be returned in order.
        if self._has_uids:
            id_fields = ["uid", "desc_uid"]
        else:
            id_fields = ["halo_id", "desc_id"]
        fields = \
          [self.field_info.resolve_field_dependencies([field])[0][0]
           for field in id_fields]
        halo_id_f, desc_id_f = fields
        dtypes = dict((field, np.int64) for field in fields)
        uid = 0
        trees = []
        nfiles = len(self.data_files)
        descs = lastids = None
        pbar = get_pbar("Planting trees", len(self.data_files))
        for i, dfl in enumerate(self.data_files):
            if not isinstance(dfl, list):
                dfl = [dfl]

            batches = []
            bsize = []
            hids = []
            ancs = defaultdict(list)
            for data_file in dfl:
                data = data_file._read_fields(fields, dtypes=dtypes)
                nhalos = len(data[halo_id_f])
                batch = np.empty(nhalos, dtype=object)

                for it in range(nhalos):
                    descid = data[desc_id_f][it]
                    if self._has_uids:
                        my_uid = data[halo_id_f][it]
                    else:
                        my_uid = uid
                    root = i == 0 or descid == -1
                    # The data says a descendent exists, but it's not there.
                    # This shouldn't happen, but it does sometimes.
                    # This can also happen when a descendent is more than
                    # one snapshot removed.
                    mcollect = False
                    if not root and descid not in lastids:
                        root = True
                        my_descid = descid
                        descid = data[desc_id_f][it] = -1
                        if self._has_uids:
                            mcollect = True
                    tree_node = TreeNode(my_uid, arbor=self, root=root)
                    tree_node._fi = it
                    tree_node.data_file = data_file
                    batch[it] = tree_node

                    if self._has_uids:
                        all_dict[my_uid] = tree_node

                    if root:
                        if mcollect:
                            tree_node._desc_uid = my_descid
                            missed_connections.append(tree_node)
                        else:
                            trees.append(tree_node)
                    else:
                        ancs[descid].append(tree_node)
                    uid += 1
                data_file.trees = batch
                batches.append(batch)
                bsize.append(batch.size)
                hids.append(data[halo_id_f])

            if i > 0:
                for descid, ancestors in ancs.items():
                    # this will not be fast
                    descendent = descs[descid == lastids][0]
                    descendent._ancestors = ancestors
                    for ancestor in ancestors:
                        ancestor._descendent = descendent

            if i < nfiles - 1:
                descs = np.empty(sum(bsize), dtype=object)
                lastids = np.empty(descs.size, dtype=np.int64)
                ib = 0
                for batch, hid, bs in zip(batches, hids, bsize):
                    descs[ib:ib+bs] = batch
                    lastids[ib:ib+bs] = hid
                    ib += bs
            pbar.update(i+1)
        pbar.finish()

        if self._has_uids:
            for node in missed_connections:
                my_desc_uid = node._desc_uid
                my_desc = all_dict[my_desc_uid]
                delattr(node, "_desc_uid")

                node._descendent = my_desc
                node.root = my_desc.root
                if my_desc._ancestors is None:
                    my_desc._ancestors = []
                my_desc._ancestors.append(node)

        self._trees = np.array(trees)
        self._size = self._trees.size

    def _setup_tree(self, tree_node):
        """
        Walk the tree and place all nodes into an array.

        This is required for field access.
        """

        if self.is_setup(tree_node):
            return

        nodes     = []
        uids      = []
        desc_uids = [-1]
        # This is redundant, but enables functionality that uses
        # the link storage, like TreeNode.get_node.
        links     = []
        for i, node in enumerate(tree_node._tree_nodes):
            node._tree_id = i
            node.root     = tree_node
            nodes.append(node)
            uids.append(node.uid)
            link = NodeLink(i)
            links.append(link)
            if i > 0:
                desc_uids.append(node.descendent.uid)
                desc_link = links[node.descendent.tree_id]
                desc_link.add_ancestor(link)
        tree_node._nodes     = np.array(nodes)
        tree_node._uids      = np.array(uids)
        tree_node._desc_uids = np.array(desc_uids)
        tree_node._tree_size = tree_node._uids.size
        tree_node._link_storage = np.array(links)
        # This should bypass any attempt to get this field in
        # the conventional way.
        if self.field_info["uid"].get("source") == "arbor":
            tree_node.field_data["uid"] = tree_node._uids
            tree_node.field_data["desc_uid"] = tree_node._desc_uids

    def _grow_tree(self, tree_node):
        """
        Trees are grown when they are planted.
        """
        pass



"""
Arbor field-related classes



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import numpy as np
import re
import weakref

from ytree.data_structures.detection import FieldDetector
from ytree.utilities.exceptions import \
    ArborFieldAlreadyExists, \
    ArborFieldCircularDependency, \
    ArborFieldDependencyNotFound, \
    ArborDerivedFieldException, \
    ArborFieldNotFound
from ytree.utilities.logger import \
    ytreeLogger as mylog

def _redshift(field, data):
    return 1. / data["scale_factor"] - 1.

def _time(field, data):
    co = data.arbor.cosmology
    if co is None:
        raise ArborDerivedFieldException(field["name"], data.arbor)
    return data.arbor.cosmology.t_from_z(data["redshift"])

def _vector_func(field, data):
    name = field["name"]
    cfields = field.get("vector_components", [f"{name}_{ax}" for ax in "xyz"])
    field_data = data.arbor.arr([data[cfield] for cfield in cfields])
    field_data = np.rollaxis(field_data, 1)
    return field_data

def _magnitude_func(field, data):
    name = field["name"][:-len("_magnitude")]
    return np.sqrt((data[name]**2).sum(axis=1))

class FieldInfoContainer(dict):
    """
    A container for information about fields.
    """

    alias_fields = ()
    known_fields = ()
    vector_fields = ("position", "velocity", "angular_momentum")
    data_types = ()

    def __init__(self, arbor):
        self.arbor = weakref.proxy(arbor)
        self._data_types = dict(self.data_types)

    def setup_known_fields(self):
        """
        Add units for fields on disk as defined in the known_fields
        tuple.
        """
        for field, units in self.known_fields:
            if field not in self:
                continue
            funits = self[field].get("units")
            if funits is None:
                self[field]["units"] = units

    def add_analysis_field(self, name, units, dtype=None, default=0):
        """
        Add an analysis field.
        """

        if name in self:
            raise ArborFieldAlreadyExists(name, arbor=self)

        if dtype is None:
            dtype = self.arbor._default_dtype

        self.arbor.analysis_field_list.append(name)
        self[name] = {"type": "analysis",
                      "default": default,
                      "dtype": dtype,
                      "units": units}

    def add_alias_field(self, alias, field, units=None,
                        force_add=True):
        """
        Add an alias field.
        """

        if alias in self:
            if force_add:
                ftype = self[alias].get("type", "on-disk")
                if ftype in ["alias", "derived"]:
                    fl = self.arbor.derived_field_list
                else:
                    fl = self.arbor.field_list
                mylog.warning(
                    f"Overriding field \"{alias}\" that already "
                    f"exists as {ftype} field.")
                fl.pop(fl.index(alias))
            else:
                return

        if field not in self:
            if force_add:
                raise ArborFieldDependencyNotFound(
                    field, alias, arbor=self)
            else:
                return

        if units is None:
            units = self[field].get("units")
        self.arbor.derived_field_list.append(alias)
        self[alias] = \
          {"type": "alias", "units": units,
           "dependencies": [field]}
        if "aliases" not in self[field]:
            self[field]["aliases"] = []
            self[field]["aliases"].append(alias)

    def setup_aliases(self):
        """
        Add aliases defined in the alias_fields tuple for each frontend.
        """
        for alias in self.alias_fields:
            aliasname, fieldname, units = alias
            if not isinstance(fieldname, tuple):
                fieldname = (fieldname,)
            for fname in fieldname:
                self.arbor.add_alias_field(
                    aliasname, fname, units=units,
                    force_add=False)

        # Fields with "/" in the name don't play well with hdf5.
        for field in self.arbor.field_list:
            if "/" not in field:
                continue
            alias = field.replace("/", "_")
            self.arbor.add_alias_field(alias, field)

    def add_derived_field(self, name, function,
                          units=None, dtype=None, description=None,
                          vector_components=None, force_add=True):
        """
        Add a derived field.
        """

        if name in self:
            if force_add:
                ftype = self[name].get("type", "on-disk")
                if ftype in ["alias", "derived"]:
                    fl = self.arbor.derived_field_list
                else:
                    fl = self.arbor.field_list
                mylog.warning(
                    f"Overriding field \"{name}\" that already "
                    f"exists as {ftype} field.")
                fl.pop(fl.index(name))
            else:
                return

        if units is None:
            units = ""
        if dtype is None:
            dtype = self.arbor._default_dtype
        info = {"name": name,
                "type": "derived",
                "function": function,
                "units": units,
                "dtype": dtype,
                "description": description}

        if vector_components is not None:
            info["vector_components"] = vector_components

        fc = FieldDetector(self.arbor, name=name)
        try:
            rv = function(info, fc)
        except ArborDerivedFieldException:
            return
        except TypeError as e:
            raise RuntimeError(
"""

Field function syntax in ytree has changed. Field functions must
now take two arguments, as in the following:
def my_field(field, data):
    return data['mass']

Check the TypeError exception above for more details.
""")
            raise e

        except ArborFieldDependencyNotFound as e:
            if force_add:
                raise e
            else:
                return

        rv.convert_to_units(units)
        info["dependencies"] = list(fc.keys())

        self.arbor.derived_field_list.append(name)
        self[name] = info

    def setup_derived_fields(self):
        """
        Add stock derived fields.
        """
        self.arbor.add_derived_field(
            "redshift", _redshift, units="", force_add=False)

        self.arbor.add_derived_field(
            "time", _time, units="Myr", force_add=False)

    def add_vector_field(self, fieldname, cfields=None):
        """
        Add vector and magnitude fields for a field with
        x/y/z components.

        If component fields not provided (with cfields kwarg), assume a
        naming convention of <fieldname>_<xyz>.
        """

        if cfields is None:
            cfields = [f"{fieldname}_{ax}" for ax in "xyz"]

        exists = all([field in self for field in cfields])
        if not exists:
            return None

        for field in cfields:
            self[field]["vector_fieldname"] = fieldname

        units = self[cfields[0]].get("units", None)
        self.arbor.add_derived_field(
            fieldname, _vector_func, vector_components=cfields, units=units)
        self.arbor.add_derived_field(
            f"{fieldname}_magnitude", _magnitude_func, units=units)
        self.vector_fields += (fieldname,)
        return fieldname

    def setup_vector_fields(self):
        """
        Add vector and magnitude fields.
        """

        vfields = set()
        candidates = defaultdict(dict)
        # match anything, followed by one xyz (case-insensitive),
        # optionally followed by more characters.
        vreg = re.compile(r"^(.+)([xyzXYZ]{1})(.*)$")

        for field in self:
            match = vreg.match(field)
            if match is None:
                continue

            pre, comp, suf = match.groups()
            # remove any trailing hyphens/underscores
            pre = re.sub("[-_]{1}$", "", pre)
            vfield = pre + suf
            candidates[vfield][comp.lower()] = field

        for vfield in candidates:
            if len(candidates[vfield]) == 3:
                vfields.add(vfield)

        added_fields = []
        for field in vfields.union(self.vector_fields):
            comp_fields = sorted(candidates[field].values()) \
              if field in candidates else None
            field = self.add_vector_field(field, cfields=comp_fields)
            if field is None:
                continue
            added_fields.append(field)

        self.vector_fields = tuple(added_fields)

    def resolve_field_dependencies(self, fields, fcache=None, fsize=None):
        """
        Divide fields into those to be read and those to generate.
        """
        if fcache is None:
            fcache = {}
        fields_to_read = []
        fields_to_generate = []
        fields_to_resolve = fields[:]

        while len(fields_to_resolve) > 0:
            field = fields_to_resolve.pop(0)
            if field in fcache:
                # Check that the field array is the size we want.
                # It might not be if it was previously gotten just
                # for the root and now we want it for the whole tree.
                if fsize is None or fcache[field].shape[0] == fsize:
                    continue
                del fcache[field]

            if field not in self:
                raise ArborFieldNotFound(field, self.arbor)

            ftype = self[field].get("type")
            if ftype == "derived" or ftype == "alias":
                deps = self[field]["dependencies"]
                if field in deps:
                    raise ArborFieldCircularDependency(
                        field, self.arbor)
                fields_to_resolve.extend(
                    set(deps).difference(set(fields_to_resolve)))
                if field not in fields_to_generate:
                    fields_to_generate.append(field)
            elif ftype == "analysis":
                fields_to_generate.append(field)
            else:
                if field not in fields_to_read:
                    fields_to_read.append(field)

        return fields_to_read, fields_to_generate

class FieldContainer(dict):
    """
    A container for field data.
    """
    def __init__(self, arbor):
        self.arbor = weakref.proxy(arbor)



"""
MoriaArbor fields



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import re

from ytree.data_structures.fields import \
    FieldInfoContainer

a_unit = "Msun/h/yr"
m_unit = "Msun/h"
p_unit = "Mpc/h"
r_unit = "kpc/h"
v_unit = "km/s"
j_unit = "Msun * Mpc * km * s**-1 * h**-2"

def _arbor_index_field(field, data):
    vals = getattr(data.arbor, f"_{field['name']}s")
    return vals[data["snap_index"].astype(int)]

class MoriaFieldInfo(FieldInfoContainer):
    alias_fields = (
        ("position_x", "x_0", p_unit),
        ("position_y", "x_1", p_unit),
        ("position_z", "x_2", p_unit),
        ("velocity_x", "v_0", v_unit),
        ("velocity_y", "v_1", v_unit),
        ("velocity_z", "v_2", v_unit),
        ("angular_momentum_x", "Jx", j_unit),
        ("angular_momentum_y", "Jy", j_unit),
        ("angular_momentum_z", "Jz", j_unit),
        ("velocity_dispersion", "vrms", v_unit),
        ("uid", "id", ""),
        ("desc_uid", "descendant_id", ""),
    )

    known_fields = (
        ("A[x]", r_unit),
        ("A[x](500c)", r_unit),
        ("A[y]", r_unit),
        ("A[y](500c)", r_unit),
        ("A[z]", r_unit),
        ("A[z](500c)", r_unit),
        ("Acc_Rate_1*Tdyn", a_unit),
        ("Acc_Rate_100Myr", a_unit),
        ("Acc_Rate_2*Tdyn", a_unit),
        ("Acc_Rate_Inst", a_unit),
        ("Acc_Rate_Mpeak", a_unit),
        ("First_Acc_Mvir", m_unit),
        ("First_Acc_Vmax", v_unit),
        ("Halfmass_Radius", r_unit),
        ("Jx", j_unit),
        ("Jy", j_unit),
        ("Jz", j_unit),
        ("M200c_all_spa", m_unit),
        ("M200c_bnd_cat", m_unit),
        ("M200c_tcr_spa", m_unit),
        ("M200m_all_spa", m_unit),
        ("M200m_all_spa_internal", m_unit),
        ("M200m_bnd_cat", m_unit),
        ("M200m_peak_cat", m_unit),
        ("M200m_tcr_spa", m_unit),
        ("M500c_all_spa", m_unit),
        ("M500c_bnd_cat", m_unit),
        ("M500c_tcr_spa", m_unit),
        ("Macc", m_unit),
        ("Mpeak", m_unit),
        ("Msp-apr-mn", m_unit),
        ("Msp-apr-p50", m_unit),
        ("Msp-apr-p70", m_unit),
        ("Msp-apr-p75", m_unit),
        ("Msp-apr-p80", m_unit),
        ("Msp-apr-p85", m_unit),
        ("Msp-apr-p90", m_unit),
        ("Mvir_all_spa", m_unit),
        ("Mvir_bnd_cat", m_unit),
        ("Mvir_tcr_spa", m_unit),
        ("R200c_all_spa", r_unit),
        ("R200c_bnd_cat", r_unit),
        ("R200c_tcr_spa", r_unit),
        ("R200m_all_spa", r_unit),
        ("R200m_all_spa_internal", r_unit),
        ("R200m_bnd_cat", r_unit),
        ("R200m_tcr_spa", r_unit),
        ("R500c_all_spa", r_unit),
        ("R500c_bnd_cat", r_unit),
        ("R500c_tcr_spa", r_unit),
        ("Rs_Klypin", r_unit),
        ("Rsp-apr-mn", r_unit),
        ("Rsp-apr-p50", r_unit),
        ("Rsp-apr-p70", r_unit),
        ("Rsp-apr-p75", r_unit),
        ("Rsp-apr-p80", r_unit),
        ("Rsp-apr-p85", r_unit),
        ("Rsp-apr-p90", r_unit),
        ("Rvir_all_spa", r_unit),
        ("Rvir_bnd_cat", r_unit),
        ("Rvir_tcr_spa", r_unit),
        ("Time_to_future_merger", "Gyr"),
        ("Vacc", v_unit),
        ("Vmax@Mpeak", v_unit),
        ("Voff", v_unit),
        ("Vpeak", v_unit),
        ("Xoff", r_unit),
        ("rs", r_unit),
        ("v", v_unit),
        ("vmax", v_unit),
        ("vrms", v_unit),
        ("x", p_unit),
    )

    def setup_known_fields(self):
        """
        Add units for all <fieldname>_<number> fields as well.
        """

        kfields = dict(self.known_fields)
        freg = re.compile(r"(^.+)_\d+$")
        for field in self:
            if self[field].get("units") is not None:
                continue

            if field in kfields:
                self[field]["units"] = kfields[field]
                continue

            fs = freg.search(field)
            if fs and fs.groups()[0] in kfields:
                self[field]["units"] = kfields[fs.groups()[0]]

    def setup_derived_fields(self):
        self.arbor.add_derived_field(
            "redshift", _arbor_index_field,
            units="", force_add=False)
        # this will add it to field list for saving
        self.arbor.field_list.append("redshift")

        self.arbor.add_derived_field(
            "scale_factor", _arbor_index_field,
            units="", force_add=False)
        # this will add it to field list for saving
        self.arbor.field_list.append("scale_factor")

        self.arbor.add_derived_field(
            "time", _arbor_index_field,
            units="Gyr", force_add=False)
        # this will add it to field list for saving
        self.arbor.field_list.append("time")

        super().setup_derived_fields()



"""
ConsistentTreesHDF5Arbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import h5py
import numpy as np
import re

from yt.funcs import \
    get_pbar

from ytree.data_structures.arbor import \
    SegmentedArbor
from ytree.frontends.consistent_trees.utilities import \
    parse_ctrees_header
from ytree.frontends.consistent_trees_hdf5.fields import \
    ConsistentTreesHDF5FieldInfo
from ytree.frontends.consistent_trees_hdf5.io import \
    ConsistentTreesHDF5DataFile, \
    ConsistentTreesHDF5RootFieldIO, \
    ConsistentTreesHDF5TreeFieldIO
from ytree.utilities.exceptions import \
    ArborDataFileEmpty
from ytree.utilities.logger import \
    ytreeLogger as mylog

_access_names = {
    'tree':   {'group'     : 'TreeInfo',
               'host_id'   : 'TreeRootID',
               'offset'    : 'TreeHalosOffset',
               'size'      : 'TreeNhalos',
               'total'     : 'TotNtrees',
               'file_size' : 'Ntrees',
               'host_attr' : 'tree_root_id'},
    'forest': {'group'     : 'ForestInfo',
               'host_id'   : 'ForestID',
               'offset'    : 'ForestHalosOffset',
               'size'      : 'ForestNhalos',
               'total'     : 'TotNforests',
               'file_size' : 'Nforests',
               'host_attr' : 'forest_id'}
}

class ConsistentTreesHDF5Arbor(SegmentedArbor):
    """
    Arbors loaded from consistent-trees data converted into HDF5.
    """

    _parameter_file_is_data_file = True
    _field_info_class = ConsistentTreesHDF5FieldInfo
    _root_field_io_class = ConsistentTreesHDF5RootFieldIO
    _tree_field_io_class = ConsistentTreesHDF5TreeFieldIO
    _default_dtype = np.float32
    _node_io_attrs = ('_fi', '_si', '_ei')

    def __init__(self, filename, access='tree'):
        if access not in _access_names:
            raise ValueError(
                f"Invalid access value: {access}. Valid options are: {_access_names}.")
        self.access = access
        self._node_io_attrs += (_access_names[access]['host_attr'],)
        super().__init__(filename)

    def _node_io_loop_finish(self, data_file):
        data_file._field_cache.reset()
        data_file.close()

    @property
    def _virtual_dataset(self):
        return re.search(r"\_\d+\.h5$", self.parameter_filename) is None

    def _get_data_files(self):
        aname = _access_names[self.access]['file_size']

        if self._virtual_dataset:
            with h5py.File(self.filename, mode='r') as f:
                self.data_files = \
                  [ConsistentTreesHDF5DataFile(self.filename, lname) for lname in f]
                self._file_count = \
                  np.array([f[lname].attrs[aname] for lname in f])
        else:
            if not isinstance(self.filename, list):
                fns = [self.filename]
            else:
                fns = self.filename
            self.data_files = [ConsistentTreesHDF5DataFile(fn, None) for fn in fns]
            self._file_count = \
              np.array([h5py.File(fn, mode='r').attrs[aname] for fn in fns])
            self._size = sum(self._file_count)

    def _parse_parameter_file(self):
        f = h5py.File(self.parameter_filename, mode='r')

        # Is the file a collection of virtual data sets
        # pointing to multiple data files?
        virtual = self._virtual_dataset
        if virtual:
            fgroup = f.get('File0')
            if fgroup is None:
                raise ArborDataFileEmpty(self.filename)
        else:
            fgroup = f

        if 'halos' in fgroup['Forests']:
            # array of structs layout
            mylog.warning(
                "This dataset was written in array of structs format. "
                "Field access will be significantly slower than struct "
                "of arrays format.")
            self._aos = True
            ftypes = fgroup['Forests/halos'].dtype
            my_fi = dict((ftypes.names[i], {'dtype': ftypes[i]})
                         for i in range(len(ftypes)))
        else:
            # struct of arrays layout
            self._aos = False
            my_fi = dict((field, {'dtype': data.dtype})
                        for field, data in fgroup['Forests'].items())

        if virtual:
            aname = _access_names[self.access]['total']
            self._size = f.attrs[aname]
        header = fgroup.attrs['Consistent Trees_metadata'].astype(str)
        header = header.tolist()
        f.close()

        header_fi = parse_ctrees_header(
            self, header, ntrees_in_file=False)
        # Do some string manipulation to match the header with
        # altered names in the hdf5 file.
        new_fi = {}
        for field in header_fi:
            new_field = field
            # remove ?| characters
            new_field = re.sub(r'[?|]', '', new_field)
            # replace []/() characters with _
            new_field = re.sub(r'[\[\]\/\(\)]', '_', new_field)
            # remove leading/trailing underscores
            new_field = new_field.strip('_')
            # replace double underscore with single underscore
            new_field = new_field.replace('__', '_')

            new_fi[new_field] = header_fi[field].copy()
            if 'column' in new_fi[new_field]:
                del new_fi[new_field]['column']

        for field in my_fi:
            my_fi[field].update(new_fi.get(field, {}))

        self.field_list = list(my_fi.keys())
        self.field_info.update(my_fi)

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        my_access  = _access_names[self.access]
        groupname  = my_access['group']
        hostname   = my_access['host_id']
        hostattr   = my_access['host_attr']
        offsetname = my_access['offset']
        sizename   = my_access['size']

        c = 0
        file_offsets = self._file_count.cumsum() - self._file_count
        pbar = get_pbar(f'Planting {self.access}s', self._size)
        for idf, data_file in enumerate(self.data_files):
            data_file.open()
            hostids = data_file.fh[groupname][hostname][()]
            offsets = data_file.fh[groupname][offsetname][()]
            tree_sizes = data_file.fh[groupname][sizename][()]
            data_file.close()

            istart = file_offsets[idf]
            iend = istart + offsets.size
            self._node_info[hostattr][istart:iend] = hostids
            self._node_info['_fi'][istart:iend] = idf
            self._node_info['_si'][istart:iend] = offsets
            self._node_info['_ei'][istart:iend] = offsets + tree_sizes
            self._node_info['_tree_size'][istart:iend] = tree_sizes
            c += offsets.size
            pbar.update(c)
        pbar.finish()
        self._node_info['uid'] = self['uid']

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        Should be an hdf5 file with a few key attributes.
        """
        fns = args[0]
        if not isinstance(fns, list):
            fns = [fns]

        for fn in fns:
            if not h5py.is_hdf5(fn):
                return False
            # single data file
            if re.search(r"\_\d+\.h5$", fn):
                attrs = ['Nforests', 'Ntrees', 'Nhalos']
            # virtual data set file
            else:
                if len(fns) > 1:
                    raise RuntimeError(
                        'Virtual data set file cannot be given in a list.')
                attrs = ['Nfiles', 'TotNforests', 'TotNhalos', 'TotNtrees']

            with h5py.File(fn, mode='r') as f:
                for attr in attrs:
                    if attr not in f.attrs:
                        return False
        return True



"""
CSVArbor class and member functions



"""

#-----------------------------------------------------------------------------
# Copyright (c) ytree development team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from collections import defaultdict
import numpy as np

from yt.funcs import get_pbar

from ytree.data_structures.arbor import CatalogArbor
from ytree.data_structures.tree_node import TreeNode

from ytree.frontends.csv.io import CSVDataFile

from ytree.utilities.io import f_text_block

from numpy.dtypes import StringDType

field_data_types = {
    "FLOAT": float,
    "INT": int,
    "STR": StringDType(),
}

class CSVArbor(CatalogArbor):
    """
    Arbors loaded from consistent-trees tree_*.dat files.
    """

    _data_file_class = CSVDataFile
    _has_uids = True
    _default_dtype = np.float32
    _node_con_attrs = ()
    _node_io_attrs = ()
    _ancestors = None

    def __init__(self, filename, sep=","):
        self.sep = sep
        super().__init__(filename)

    def _get_data_files(self):
        self.data_files = [CSVDataFile(self.filename, self)]

    def _parse_parameter_file(self, filename=None):
        if filename is None:
            filename = self.filename

        with open(filename, mode="r") as f:
            ldata = []
            for i in range(3):
                ldata.append(f.readline().strip()[1:].split(self.sep))
            self._hoffset = f.tell()

        lens = [len(_) for _ in ldata]
        if min(lens) != max(lens):
            raise RuntimeError(
                "Header lines must have same number of values.")

        fields, dtypes, units = ldata
        fi = {}
        for i, (field, dtype, unit) in enumerate(zip(fields, dtypes, units)):
            my_unit = "" if unit == "None" else unit
            fi[field] = {"column": i, "units": my_unit,
                         "dtype": field_data_types[dtype]}

        try:
            fi["uid"]["dtype"] = np.int64
            fi["desc_uid"]["dtype"] = np.int64
        except KeyError:
            raise IOError(f"{filename} is missing either uid or desc_uid fields.")

        self.field_list = list(fi.keys())
        self.field_info.update(fi)

    def _plant_trees(self):
        if self.is_planted or self._size == 0:
            return

        data_file = self.data_files[0]
        data_file.open()
        data_file.fh.seek(self._hoffset)

        fi = self.field_info
        col_uid = fi["uid"]["column"]
        typ_uid = fi["uid"]["dtype"]
        col_des = fi["desc_uid"]["column"]
        typ_des = fi["desc_uid"]["dtype"]

        roots = []
        root_offsets = []

        # these two are for nonroots only
        desc_uids = {}
        offsets = {}

        for line, loc in f_text_block(
                data_file.fh, pbar_string="Loading tree roots"):

            online = line.split(self.sep)
            uid = typ_uid(online[col_uid])
            desc_uid = typ_des(online[col_des])

            if desc_uid == -1:
                roots.append(uid)
                root_offsets.append(loc)
            else:
                desc_uids[uid] = desc_uid
                offsets[uid] = loc

        n_nonroots = len(desc_uids)
        uids = np.array(list(desc_uids.keys()) + roots)
        nr_desc_uids = np.array(list(desc_uids.values()))
        # look for any nodes with missing descendents
        missing = np.isin(nr_desc_uids, uids, invert=True)
        # need to address only up to n_nonroots since
        # uids has everything and desc_uids has only nonroots
        missing_uids = uids[:n_nonroots][missing]
        for uid in missing_uids:
            del desc_uids[uid]

        trees = []
        for uid, offset in zip(roots, root_offsets):
            my_node = TreeNode(uid, arbor=self, root=True)
            my_node._offset = offset
            my_node.data_file = self.data_files[0]
            trees.append(my_node)

        for uid in missing_uids:
            offset = offsets.pop(uid)
            my_node = TreeNode(uid, arbor=self, root=True)
            my_node._offset = offset
            my_node.data_file = self.data_files[0]
            trees.append(my_node)

        self._size = len(trees)
        self._trees = np.array(trees)

        # these are for the nonroot nodes
        # we will construct them later when they are needed
        self._desc_uids = desc_uids
        self._offsets = offsets

    def _build_trees(self):
        """
        Resolve the dictionary of desc_uids into trees.
        """

        # If we have already done this, get out of here.
        if self._ancestors is not None:
            return

        if self.size < 1:
            return

        desc_uids = getattr(self, "_desc_uids", None)
        if desc_uids is None:
            return

        offsets = getattr(self, "_offsets")
        ancestors = defaultdict(list)
        pbar = get_pbar("Building trees", len(desc_uids))
        for i, (uid, desc_uid) in enumerate(desc_uids.items()):
            my_node = TreeNode(uid, arbor=self)
            my_node._offset = offsets[uid]
            my_node.data_file = self.data_files[0]
            ancestors[desc_uid].append(my_node)
            pbar.update(i+1)
        pbar.finish()

        self._ancestors = ancestors
        del self._desc_uids, self._offsets

    def _grow_tree(self, tree_node):
        """
        We don't actually attach the tree's _ancestors during planting,
        so do it here.
        """

        self._get_ancestors(tree_node)

    def _get_ancestors(self, tree_node):
        """
        Get this tree's ancestors from the main ancestor dictionary.

        This arbor stores every tree's ancestors in a dictionary.
        When we finally need it, grab it from the arbor and attach
        it to the tree.
        """

        self._build_trees()
        if tree_node._ancestors is None:
            tree_node._ancestors = self._ancestors.pop(tree_node.uid, [])
        for node in tree_node._ancestors:
            node._descendent = tree_node
            self._get_ancestors(node)

    def _setup_tree(self, tree_node):
        if self.is_setup(tree_node):
            return

        self._get_ancestors(tree_node)
        super()._setup_tree(tree_node)

    @classmethod
    def _is_valid(self, *args, **kwargs):
        """
        File should end in .csv and start with three lines of the
        following format.
        #<fieldname>,<fieldname>,<fieldname>
        #<type>,<type>,<type>
        #<units>,<units>,<units>

        For example:
        #uid,desc_uid,mass,redshift
        #int,int,float,float
        #None,None,"Msun",None
        """
        fn = args[0]
        sep = kwargs.get("sep", ",")
        if not fn.endswith(".csv"):
            return False
        with open(fn, "r") as f:
            for i in range(3):
                line = f.readline()
                if not line.startswith("#"):
                    return False

                if sep not in line:
                    return False
        return True



from ytree.frontends.csv import \
    CSVArbor
from ytree.utilities.testing import \
    ArborTest, \
    TempDirTest

class CSVArborTest(TempDirTest, ArborTest):
    arbor_type = CSVArbor
    test_filename = "csv/trees.csv"

    def load_callback(self, a):
        a.set_selector("max_field_value", "charisma")
